import warnings
from datetime import datetime
from typing import Dict, List, Optional, Tuple

import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy
import statsmodels.api as sm
import ujson as json
from scipy.optimize import minimize
from sklearn.linear_model import LinearRegression
from termcolor import colored

from CurvyCUSIPs.CurveBuilder import calc_ust_impl_spot_n_fwd_curve, calc_ust_metrics
from CurvyCUSIPs.CurveDataFetcher import CurveDataFetcher
from CurvyCUSIPs.utils.arbitragelab import EngleGrangerPortfolio, JohansenPortfolio, construct_spread
from CurvyCUSIPs.utils.regression_utils import run_odr


def dv01_neutral_curve_hegde_ratio(
    as_of_date: datetime,
    front_leg_bond_row: Dict | pd.Series,
    back_leg_bond_row: Dict | pd.Series,
    curve_data_fetcher: CurveDataFetcher,
    scipy_interp_curve: scipy.interpolate.interpolate,
    repo_rate: float,
    quote_type: Optional[str] = "eod",
    front_leg_par_amount: Optional[int] = None,
    back_leg_par_amount: Optional[int] = None,
    total_trade_par_amount: Optional[int] = None,
    yvx_beta_adjustment: Optional[int] = None,
    verbose: Optional[bool] = True,
    very_verbose: Optional[bool] = False,
):
    if isinstance(front_leg_bond_row, pd.Series) or isinstance(front_leg_bond_row, pd.DataFrame):
        front_leg_bond_row = front_leg_bond_row.to_dict("records")[0]
    if isinstance(back_leg_bond_row, pd.Series) or isinstance(back_leg_bond_row, pd.DataFrame):
        back_leg_bond_row = back_leg_bond_row.to_dict("records")[0]

    front_leg_info = curve_data_fetcher.ust_data_fetcher.cusip_to_ust_label(cusip=front_leg_bond_row["cusip"])
    back_leg_info = curve_data_fetcher.ust_data_fetcher.cusip_to_ust_label(cusip=back_leg_bond_row["cusip"])

    front_leg_metrics = calc_ust_metrics(
        bond_info=front_leg_info,
        curr_price=front_leg_bond_row[f"{quote_type}_price"],
        curr_ytm=front_leg_bond_row[f"{quote_type}_yield"],
        as_of_date=as_of_date,
        scipy_interp=scipy_interp_curve,
        on_rate=repo_rate,
    )
    back_leg_metrics = calc_ust_metrics(
        bond_info=back_leg_info,
        curr_price=back_leg_bond_row[f"{quote_type}_price"],
        curr_ytm=back_leg_bond_row[f"{quote_type}_yield"],
        as_of_date=as_of_date,
        scipy_interp=scipy_interp_curve,
        on_rate=repo_rate,
    )

    front_leg_ttm: float = (front_leg_info["maturity_date"] - as_of_date).days / 365
    back_leg_ttm: float = (back_leg_info["maturity_date"] - as_of_date).days / 365
    impl_spot_3m_fwds = calc_ust_impl_spot_n_fwd_curve(n=0.25, scipy_interp_curve=scipy_interp_curve, return_scipy=True)
    impl_spot_6m_fwds = calc_ust_impl_spot_n_fwd_curve(n=0.5, scipy_interp_curve=scipy_interp_curve, return_scipy=True)
    impl_spot_12m_fwds = calc_ust_impl_spot_n_fwd_curve(n=1, scipy_interp_curve=scipy_interp_curve, return_scipy=True)

    if very_verbose:
        print("Front Leg Info: ")
        print(front_leg_bond_row)
        print(front_leg_metrics)
        print("Back Leg Info: ")
        print(back_leg_bond_row)
        print(back_leg_metrics)

    if front_leg_bond_row["rank"] == 0 and back_leg_bond_row["rank"] == 0:
        print(f"{front_leg_bond_row["original_security_term"].split("-")[0]}s{back_leg_bond_row["original_security_term"].split("-")[0]}s")
    print(f"{front_leg_bond_row["ust_label"]} / {back_leg_bond_row["ust_label"]}") if verbose else None

    hr = back_leg_metrics["bps"] / front_leg_metrics["bps"]
    print(colored(f"BPV Neutral Hedge Ratio: {hr}", "light_blue")) if verbose else None
    if yvx_beta_adjustment:
        (print(colored(f"Beta Weighted Hedge Ratio: {hr * yvx_beta_adjustment:3f}", "light_magenta")) if verbose else None)
        hr = hr * yvx_beta_adjustment

    if total_trade_par_amount is not None:
        if front_leg_par_amount is not None or back_leg_par_amount is not None:
            raise ValueError("Cannot provide total_trade_par_amount along with individual leg par amounts.")

        if hr <= 0:
            raise ValueError("Hedge ratio must be positive to calculate par amounts.")

        front_leg_par_amount = total_trade_par_amount / (hr + 1)
        back_leg_par_amount = hr * front_leg_par_amount

    else:
        if front_leg_par_amount and back_leg_par_amount:
            raise ValueError("'front_leg_par_amount' and 'back_leg_par_amount' are both defined!")
        if not front_leg_par_amount and not back_leg_par_amount:
            back_leg_par_amount = 1_000_000
        if back_leg_par_amount:
            front_leg_par_amount = back_leg_par_amount * hr
        elif front_leg_par_amount:
            back_leg_par_amount = front_leg_par_amount / hr

    if verbose:
        print(
            f"Front Leg: {front_leg_bond_row["ust_label"]} (OST {front_leg_bond_row["original_security_term"]}, TTM = {front_leg_bond_row["time_to_maturity"]:3f}) Par Amount = {front_leg_par_amount :_}"
        )
        print(
            f"Back Leg: {back_leg_bond_row["ust_label"]} (OST {back_leg_bond_row["original_security_term"]}, TTM = {back_leg_bond_row["time_to_maturity"]:3f}) Par Amount = {back_leg_par_amount:_}"
        )
        print(f"Total Trade Par Amount: {front_leg_par_amount + back_leg_par_amount:_}")
        risk_weight = (front_leg_par_amount * front_leg_metrics["bps"] / 100) / (back_leg_par_amount * back_leg_metrics["bps"] / 100)
        print(f"Risk Weights: {risk_weight:3f} : 100")

    return {
        "current_spread": (back_leg_bond_row[f"{quote_type}_yield"] - front_leg_bond_row[f"{quote_type}_yield"]) * 100,
        "current_bpv_neutral_spread": (
            back_leg_bond_row[f"{quote_type}_yield"]
            - (front_leg_bond_row[f"{quote_type}_yield"] * (back_leg_metrics["bps"] / front_leg_metrics["bps"]))
        )
        * 100,
        "current_beta_weighted_spread": (
            (back_leg_bond_row[f"{quote_type}_yield"] - (front_leg_bond_row[f"{quote_type}_yield"] * hr)) * 100 if yvx_beta_adjustment else None
        ),
        "rough_3m_impl_fwd_spread": (impl_spot_3m_fwds(back_leg_ttm) - impl_spot_3m_fwds(front_leg_ttm)) * 100,
        "rough_6m_impl_fwd_spread": (impl_spot_6m_fwds(back_leg_ttm) - impl_spot_6m_fwds(front_leg_ttm)) * 100,
        "rough_12m_impl_fwd_spread": (impl_spot_12m_fwds(back_leg_ttm) - impl_spot_12m_fwds(front_leg_ttm)) * 100,
        "front_leg_metrics": front_leg_metrics,
        "back_leg_metrics": back_leg_metrics,
        "bpv_hedge_ratio": back_leg_metrics["bps"] / front_leg_metrics["bps"],
        "beta_weighted_hedge_ratio": ((back_leg_metrics["bps"] / front_leg_metrics["bps"]) * yvx_beta_adjustment if yvx_beta_adjustment else None),
        "front_leg_par_amount": front_leg_par_amount,
        "back_leg_par_amount": back_leg_par_amount,
        "spread_dv01": np.abs(back_leg_metrics["bps"] * back_leg_par_amount / 100),
        "rough_3m_carry_roll": (back_leg_metrics["rough_carry"] + back_leg_metrics["rough_3m_rolldown"])
        - hr * (front_leg_metrics["rough_carry"] + front_leg_metrics["rough_3m_rolldown"]),
        "rough_6m_carry_roll": (back_leg_metrics["rough_carry"] + back_leg_metrics["rough_6m_rolldown"])
        - hr * (front_leg_metrics["rough_carry"] + front_leg_metrics["rough_6m_rolldown"]),
        "rough_12m_carry_roll": (back_leg_metrics["rough_carry"] + back_leg_metrics["rough_12m_rolldown"])
        - hr * (front_leg_metrics["rough_carry"] + front_leg_metrics["rough_12m_rolldown"]),
    }


# reference point is buying the belly => fly spread down
def dv01_neutral_butterfly_hegde_ratio(
    as_of_date: datetime,
    front_wing_bond_row: Dict | pd.Series,
    belly_bond_row: Dict | pd.Series,
    back_wing_bond_row: Dict | pd.Series,
    curve_data_fetcher: CurveDataFetcher,
    scipy_interp_curve: scipy.interpolate.interpolate,
    repo_rate: float,
    quote_type: Optional[str] = "eod",
    front_wing_par_amount: Optional[int] = None,
    belly_par_amount: Optional[int] = None,
    back_wing_par_amount: Optional[int] = None,
    total_trade_par_amount: Optional[int] = None,
    yvx_front_wing_beta_adjustment: Optional[int] = None,
    yvx_back_wing_beta_adjustment: Optional[int] = None,
    verbose: Optional[bool] = True,
    very_verbose: Optional[bool] = False,
):
    if isinstance(front_wing_bond_row, pd.Series) or isinstance(front_wing_bond_row, pd.DataFrame):
        front_wing_bond_row = front_wing_bond_row.to_dict("records")[0]
    if isinstance(belly_bond_row, pd.Series) or isinstance(belly_bond_row, pd.DataFrame):
        belly_bond_row = belly_bond_row.to_dict("records")[0]
    if isinstance(back_wing_bond_row, pd.Series) or isinstance(back_wing_bond_row, pd.DataFrame):
        back_wing_bond_row = back_wing_bond_row.to_dict("records")[0]

    front_wing_info = curve_data_fetcher.ust_data_fetcher.cusip_to_ust_label(cusip=front_wing_bond_row["cusip"])
    belly_info = curve_data_fetcher.ust_data_fetcher.cusip_to_ust_label(cusip=belly_bond_row["cusip"])
    back_wing_info = curve_data_fetcher.ust_data_fetcher.cusip_to_ust_label(cusip=back_wing_bond_row["cusip"])

    front_wing_metrics = calc_ust_metrics(
        bond_info=front_wing_info,
        curr_price=front_wing_bond_row[f"{quote_type}_price"],
        curr_ytm=front_wing_bond_row[f"{quote_type}_yield"],
        as_of_date=as_of_date,
        scipy_interp=scipy_interp_curve,
        on_rate=repo_rate,
    )
    belly_metrics = calc_ust_metrics(
        bond_info=belly_info,
        curr_price=belly_bond_row[f"{quote_type}_price"],
        curr_ytm=belly_bond_row[f"{quote_type}_yield"],
        as_of_date=as_of_date,
        scipy_interp=scipy_interp_curve,
        on_rate=repo_rate,
    )
    back_wing_metrics = calc_ust_metrics(
        bond_info=back_wing_info,
        curr_price=back_wing_bond_row[f"{quote_type}_price"],
        curr_ytm=back_wing_bond_row[f"{quote_type}_yield"],
        as_of_date=as_of_date,
        scipy_interp=scipy_interp_curve,
        on_rate=repo_rate,
    )

    front_wing_ttm: float = (front_wing_info["maturity_date"] - as_of_date).days / 365
    belly_ttm: float = (belly_info["maturity_date"] - as_of_date).days / 365
    back_wing_ttm: float = (back_wing_info["maturity_date"] - as_of_date).days / 365
    impl_spot_3m_fwds = calc_ust_impl_spot_n_fwd_curve(n=0.25, scipy_interp_curve=scipy_interp_curve, return_scipy=True)
    impl_spot_6m_fwds = calc_ust_impl_spot_n_fwd_curve(n=0.5, scipy_interp_curve=scipy_interp_curve, return_scipy=True)
    impl_spot_12m_fwds = calc_ust_impl_spot_n_fwd_curve(n=1, scipy_interp_curve=scipy_interp_curve, return_scipy=True)

    if very_verbose:
        print("Front Wing Info: ")
        print(front_wing_info)
        print(front_wing_metrics)

        print("Belly Info: ")
        print(belly_info)
        print(belly_metrics)

        print("Back Wing Info: ")
        print(back_wing_info)
        print(back_wing_metrics)

    hedge_ratios = {
        "front_wing_hr": belly_metrics["bps"] / front_wing_metrics["bps"] / 2,
        "belly_hr": 1,
        "back_wing_hr": belly_metrics["bps"] / back_wing_metrics["bps"] / 2,
    }

    if verbose:
        if front_wing_bond_row["rank"] == 0 and belly_bond_row["rank"] == 0 and back_wing_bond_row["rank"] == 0:
            print(
                f"{front_wing_bond_row["original_security_term"].split("-")[0]}s{belly_bond_row["original_security_term"].split("-")[0]}s{back_wing_bond_row["original_security_term"].split("-")[0]}s"
            )

        (print(f"{front_wing_bond_row["ust_label"]} - {belly_bond_row["ust_label"]} - {back_wing_bond_row["ust_label"]} Fly") if verbose else None)
        print(colored(f"BPV Neutral Hedge Ratio:", "light_blue")) if verbose else None
        print(json.dumps(hedge_ratios, indent=4)) if verbose else None

        if yvx_front_wing_beta_adjustment and yvx_back_wing_beta_adjustment:
            print(colored(f"Beta Weighted Hedge Ratio:", "light_magenta")) if verbose else None
            hedge_ratios = {
                "front_wing_hr": (belly_metrics["bps"] / front_wing_metrics["bps"] / 2) * yvx_front_wing_beta_adjustment,
                "belly_hr": 1,
                "back_wing_hr": (belly_metrics["bps"] / back_wing_metrics["bps"] / 2) * yvx_back_wing_beta_adjustment,
            }
            print(json.dumps(hedge_ratios, indent=4)) if verbose else None

        if total_trade_par_amount is not None:
            if front_wing_par_amount is not None or belly_par_amount is not None or back_wing_par_amount is not None:
                raise ValueError("Cannot provide total_trade_par_amount along with individual leg par amounts.")

            total_hr_abs = abs(hedge_ratios["front_wing_hr"]) + abs(hedge_ratios["belly_hr"]) + abs(hedge_ratios["back_wing_hr"])
            belly_par_amount = total_trade_par_amount / total_hr_abs
            front_wing_par_amount = hedge_ratios["front_wing_hr"] * belly_par_amount
            back_wing_par_amount = hedge_ratios["back_wing_hr"] * belly_par_amount

        else:
            if belly_par_amount:
                front_wing_par_amount = hedge_ratios["front_wing_hr"] * belly_par_amount
                belly_par_amount = belly_par_amount
                back_wing_par_amount = hedge_ratios["back_wing_hr"] * belly_par_amount
            elif front_wing_par_amount:
                front_wing_par_amount = front_wing_par_amount
                belly_par_amount = front_wing_par_amount / hedge_ratios["front_wing_hr"]
                back_wing_par_amount = hedge_ratios["back_wing_hr"] * (front_wing_par_amount / hedge_ratios["front_wing_hr"])
            elif back_wing_par_amount:
                front_wing_par_amount = hedge_ratios["front_wing_hr"] * (back_wing_par_amount / hedge_ratios["back_wing_hr"])
                belly_par_amount = back_wing_par_amount / hedge_ratios["back_wing_hr"]
                back_wing_par_amount = back_wing_par_amount

        print(
            f"Front Wing: {front_wing_bond_row["ust_label"]} (OST {front_wing_bond_row["original_security_term"]}, TTM = {front_wing_bond_row["time_to_maturity"]:3f}) Par Amount = {front_wing_par_amount:_}"
        )
        print(
            f"Belly: {belly_bond_row["ust_label"]} (OST {belly_bond_row["original_security_term"]}, TTM = {belly_bond_row["time_to_maturity"]:3f}) Par Amount = {belly_par_amount:_}"
        )
        print(
            f"Back Wing: {back_wing_bond_row["ust_label"]} (OST {back_wing_bond_row["original_security_term"]}, TTM = {back_wing_bond_row["time_to_maturity"]:3f}) Par Amount = {back_wing_par_amount:_}"
        )
        print(f"Total Trade Par Amount: {front_wing_par_amount + belly_par_amount + back_wing_par_amount:_}")
        (
            print(
                f"Risk Weights - Front Wing: {yvx_front_wing_beta_adjustment:.3%}, Back Wing: {yvx_back_wing_beta_adjustment:.3%}, Sum: {yvx_front_wing_beta_adjustment + yvx_back_wing_beta_adjustment:.3%}"
            )
            if yvx_front_wing_beta_adjustment and yvx_back_wing_beta_adjustment
            else None
        )

    return {
        "curr_spread": (
            (belly_bond_row[f"{quote_type}_yield"] - front_wing_bond_row[f"{quote_type}_yield"])
            - (back_wing_bond_row[f"{quote_type}_yield"] - belly_bond_row[f"{quote_type}_yield"])
        )
        * 100,
        "rough_3m_impl_fwd_spread": (
            (impl_spot_3m_fwds(belly_ttm) - impl_spot_3m_fwds(front_wing_ttm)) - (impl_spot_3m_fwds(back_wing_ttm) - impl_spot_3m_fwds(belly_ttm))
        )
        * 100,
        "rough_6m_impl_fwd_spread": (
            (impl_spot_6m_fwds(belly_ttm) - impl_spot_6m_fwds(front_wing_ttm)) - (impl_spot_6m_fwds(back_wing_ttm) - impl_spot_6m_fwds(belly_ttm))
        )
        * 100,
        "rough_12m_impl_fwd_spread": (
            (impl_spot_12m_fwds(belly_ttm) - impl_spot_12m_fwds(front_wing_ttm)) - (impl_spot_12m_fwds(back_wing_ttm) - impl_spot_12m_fwds(belly_ttm))
        )
        * 100,
        "front_wing_metrics": front_wing_metrics,
        "belly_metrics": belly_metrics,
        "back_wing_metrics": back_wing_metrics,
        "bpv_neutral_hedge_ratio": {
            "front_wing_hr": belly_metrics["bps"] / front_wing_metrics["bps"] / 2,
            "belly_hr": 1,
            "back_wing_hr": belly_metrics["bps"] / back_wing_metrics["bps"] / 2,
        },
        "beta_weighted_hedge_ratio": (hedge_ratios if yvx_front_wing_beta_adjustment and yvx_back_wing_beta_adjustment else None),
        "front_wing_par_amount": front_wing_par_amount,
        "belly_par_amount": belly_par_amount,
        "back_leg_par_amount": back_wing_par_amount,
        "spread_dv01": np.abs(belly_metrics["bps"] * belly_par_amount / 100),
        "rough_3m_carry_roll": (belly_metrics["rough_carry"] + belly_metrics["rough_3m_rolldown"])
        - (hedge_ratios["front_wing_hr"] * (front_wing_metrics["rough_carry"] + front_wing_metrics["rough_3m_rolldown"]))
        - (hedge_ratios["back_wing_hr"] * (back_wing_metrics["rough_carry"] + back_wing_metrics["rough_3m_rolldown"])),
        "rough_6m_carry_roll": (belly_metrics["rough_carry"] + belly_metrics["rough_6m_rolldown"])
        - (hedge_ratios["front_wing_hr"] * (front_wing_metrics["rough_carry"] + front_wing_metrics["rough_6m_rolldown"]))
        - (hedge_ratios["back_wing_hr"] * (back_wing_metrics["rough_carry"] + back_wing_metrics["rough_6m_rolldown"])),
        "rough_12m_carry_roll": (belly_metrics["rough_carry"] + belly_metrics["rough_12m_rolldown"])
        - (hedge_ratios["front_wing_hr"] * (front_wing_metrics["rough_carry"] + front_wing_metrics["rough_12m_rolldown"]))
        - (hedge_ratios["back_wing_hr"] * (back_wing_metrics["rough_carry"] + back_wing_metrics["rough_12m_rolldown"])),
    }


# https://github.com/hudson-and-thames/arbitragelab/blob/master/arbitragelab/hedge_ratios/box_tiao.py
def get_box_tiao_hedge_ratio(price_data: pd.DataFrame, dependent_variable: str) -> Tuple[dict, pd.DataFrame, None, pd.Series]:
    """
    Perform Box-Tiao canonical decomposition on the assets dataframe.
    The resulting ratios are the weightings of each asset in the portfolio. There are N decompositions for N assets,
    where each column vector corresponds to one portfolio. The order of the weightings corresponds to the
    descending order of the eigenvalues.
    """

    def _least_square_VAR_fit(demeaned_price_data: pd.DataFrame) -> np.array:
        var_model = sm.tsa.VAR(demeaned_price_data)
        least_sq_est = np.squeeze(var_model.fit(1).coefs, axis=0)
        return least_sq_est, var_model

    X = price_data.copy()
    X = X[[dependent_variable] + [x for x in X.columns if x != dependent_variable]]

    demeaned = X - X.mean()
    least_sq_est, var_model = _least_square_VAR_fit(demeaned)
    covar = demeaned.cov()
    box_tiao_matrix = np.linalg.inv(covar) @ least_sq_est @ covar @ least_sq_est.T
    eigvals, eigvecs = np.linalg.eig(box_tiao_matrix)
    bt_eigvecs = eigvecs[:, np.argsort(eigvals)[::-1]]
    hedge_ratios = dict(zip(X.columns, bt_eigvecs[:, -1]))

    beta_weights = []
    for ticker, h in hedge_ratios.items():
        if ticker != dependent_variable:
            beta = -h / hedge_ratios[dependent_variable]
            hedge_ratios[ticker] = beta
            beta_weights.append(beta)
    hedge_ratios[dependent_variable] = 1.0
    residuals = construct_spread(price_data, hedge_ratios=hedge_ratios, dependent_variable=dependent_variable)

    return {
        "beta_weights": beta_weights,
        "hedge_ratios_dict": hedge_ratios,
        "X": X,
        "residuals": residuals,
        "results": var_model,
    }


# https://github.com/hudson-and-thames/arbitragelab/blob/master/arbitragelab/hedge_ratios/johansen.py
def get_johansen_hedge_ratio(price_data: pd.DataFrame, dependent_variable: str) -> Tuple[dict, pd.DataFrame, pd.Series, pd.Series]:
    """
    Get hedge ratio from Johansen test eigenvector
    https://en.wikipedia.org/wiki/Johansen_test
    """

    port = JohansenPortfolio()
    port.fit(price_data, dependent_variable)

    X = price_data.copy()
    X.drop(columns=dependent_variable, axis=1, inplace=True)

    y = price_data[dependent_variable].copy()

    hedge_ratios = port.hedge_ratios.iloc[0].to_dict()
    residuals = construct_spread(price_data, hedge_ratios=hedge_ratios, dependent_variable=dependent_variable)

    hedge_ratios_copy = hedge_ratios.copy()
    del hedge_ratios_copy[dependent_variable]

    return {
        "beta_weights": list(hedge_ratios_copy.values()),
        "hedge_ratios_dict": hedge_ratios,
        "X": X,
        "y": y,
        "residuals": residuals,
        "results": port,
    }


# https://github.com/hudson-and-thames/arbitragelab/blob/master/arbitragelab/hedge_ratios/half_life.py
def get_minimum_hl_hedge_ratio(price_data: pd.DataFrame, dependent_variable: str) -> Tuple[dict, pd.DataFrame, pd.Series, pd.Series, object]:
    """
    Get hedge ratio by minimizing spread half-life of mean reversion.
    https://quant.stackexchange.com/questions/77953/interpretation-and-intuition-behind-half-life-of-a-mean-reverting-process
    """

    def get_half_life_of_mean_reversion_ou_process(data: pd.Series) -> float:
        reg = LinearRegression(fit_intercept=True)
        training_data = data.shift(1).dropna().values.reshape(-1, 1)
        target_values = data.diff().dropna()
        reg.fit(X=training_data, y=target_values)
        half_life = -np.log(2) / reg.coef_[0]
        return half_life

    def _min_hl_function(beta: np.array, X: pd.DataFrame, y: pd.Series) -> float:
        spread = y - (beta * X).sum(axis=1)
        return abs(get_half_life_of_mean_reversion_ou_process(spread))

    X = price_data.copy()
    X.drop(columns=dependent_variable, axis=1, inplace=True)

    y = price_data[dependent_variable].copy()
    initial_guess = (y[0] / X).mean().values
    result = minimize(_min_hl_function, x0=initial_guess, method="BFGS", tol=1e-5, args=(X, y))

    if result.status != 0:
        warnings.warn("Minimum Half Life Optimization failed to converge. Please check output hedge ratio! The result can be unstable!")

    residuals = y - (result.x * X).sum(axis=1)

    hedge_ratios = result.x
    hedge_ratios_dict = dict(zip([dependent_variable] + X.columns.tolist(), np.insert(hedge_ratios, 0, 1.0)))

    return {
        "beta_weights": list(hedge_ratios),
        "hedge_ratios_dict": hedge_ratios_dict,
        "X": X,
        "y": y,
        "residuals": residuals,
        "results": result,
    }


# https://github.com/hudson-and-thames/arbitragelab/blob/master/arbitragelab/hedge_ratios/adf_optimal.py
def get_adf_optimal_hedge_ratio(price_data: pd.DataFrame, dependent_variable: str) -> Tuple[dict, pd.DataFrame, pd.Series, pd.Series, object]:
    """
    Get hedge ratio by minimizing ADF test statistic.
    https://www.statisticshowto.com/adf-augmented-dickey-fuller-test/
    """

    def _min_adf_stat(beta: np.array, X: pd.DataFrame, y: pd.Series) -> float:
        portfolio = EngleGrangerPortfolio()
        spread = y - (beta * X).sum(axis=1)
        portfolio.perform_eg_test(spread)
        return portfolio.adf_statistics.loc["statistic_value"].iloc[0]

    X = price_data.copy()
    X.drop(columns=dependent_variable, axis=1, inplace=True)
    y = price_data[dependent_variable].copy()
    initial_guess = (y[0] / X).mean().values
    result = minimize(_min_adf_stat, x0=initial_guess, method="BFGS", tol=1e-5, args=(X, y))

    if result.status != 0:
        warnings.warn("ADF Optimization failed to converge. Please check output hedge ratio! The result can be unstable!")

    residuals = y - (result.x * X).sum(axis=1)

    hedge_ratios = result.x
    hedge_ratios_dict = dict(zip([dependent_variable] + X.columns.tolist(), np.insert(hedge_ratios, 0, 1.0)))

    return {
        "beta_weights": list(hedge_ratios),
        "hedge_ratios_dict": hedge_ratios_dict,
        "X": X,
        "y": y,
        "residuals": residuals,
        "results": result,
    }


def beta_estimates(
    df: pd.DataFrame,
    x_cols: List[str],
    y_col: str,
    run_on_level_changes: Optional[bool] = False,
    x_errs: Optional[npt.ArrayLike] = None,
    y_errs: Optional[npt.ArrayLike] = None,
    pc_scores_df: Optional[pd.DataFrame] = None,
    loadings_df: Optional[pd.DataFrame] = None,
) -> Dict:
    df = df[["Date"] + x_cols + [y_col]].copy()
    df_level = df.copy()

    if len(x_cols) == 1:
        df["spread"] = df[y_col] - df[x_cols[0]]
    elif len(x_cols) == 2:
        df["spread"] = (df[y_col] - df[x_cols[0]]) - (df[x_cols[1]] - df[y_col])
    else:
        raise ValueError("Too many x_cols")

    if run_on_level_changes:
        date_col = df["Date"]
        df = df[x_cols + [y_col] + ["spread"]].diff()
        df["Date"] = date_col
    df = df.dropna()

    pc1_beta = None
    if loadings_df is not None:
        ep_x0_pc1 = loadings_df.loc[x_cols[0], "PC1"]
        ep_x0_pc2 = loadings_df.loc[x_cols[0], "PC2"]
        ep_x0_pc3 = loadings_df.loc[x_cols[0], "PC3"] if len(x_cols) > 1 else None

        ep_y_pc1 = loadings_df.loc[y_col, "PC1"]
        ep_y_pc2 = loadings_df.loc[y_col, "PC2"]
        ep_y_pc3 = loadings_df.loc[y_col, "PC3"] if len(x_cols) > 1 else None

        if len(x_cols) > 1:
            ep_x1_pc1 = loadings_df.loc[x_cols[1], "PC1"]
            ep_x1_pc2 = loadings_df.loc[x_cols[1], "PC2"]
            ep_x1_pc3 = loadings_df.loc[x_cols[1], "PC3"] if len(x_cols) > 1 else None

            # see  Doug Huggins, Christian Schaller Fixed Income Relative Value Analysis ed2 page 76 APPROPRIATE HEDGING
            r"""
                Hedge ratios against more factors are best calculated via matrix inversion. 
                For example, the hedge ratio for a 2Y-5Y-10Y butterfly which is neutral to changes in the first and 
                second factor can be calculated for a given notional $n_5$ for 5Y by:
                    $$
                        \begin{pmatrix}
                            n_2 \\
                            n_{10}
                        \end{pmatrix}
                        = 
                        \begin{pmatrix}
                            BPV_2 \cdot e_{12} & BPV_2 \cdot e_{22} \\
                            BPV_{10} \cdot e_{12} & BPV_{10} \cdot e_{22}
                            \end{pmatrix}^{-1} 
                            \begin{pmatrix}
                            - n_5 \cdot BPV_5 \cdot e_{15} \\
                            - n_5 \cdot BPV_5 \cdot e_{25}
                        \end{pmatrix}
                    $$
            """
            pc1_beta = list(
                np.dot(
                    np.linalg.inv(np.array([[ep_x0_pc1, ep_x1_pc1], [ep_x0_pc2, ep_x1_pc2]])),
                    np.array([ep_y_pc1, ep_y_pc2]),
                )
            )
        else:
            pc1_beta = ep_y_pc1 / ep_x0_pc1

    # avoiding divide by zero errors
    small_value = 1e-8
    if x_errs is not None:
        x_errs[x_errs == 0] = small_value
    if y_errs is not None:
        y_errs[y_errs == 0] = small_value

    regression_results = {
        "ols": sm.OLS(df[y_col], sm.add_constant(df[x_cols])).fit(),
        "tls": run_odr(df=df, x_cols=x_cols, y_col=y_col, x_errs=None, y_errs=None),
        # ODR becomes TLS if errors not specified
        "odr": (run_odr(df=df, x_cols=x_cols, y_col=y_col, x_errs=x_errs, y_errs=y_errs) if x_errs is not None or y_errs is not None else None),
        "box_tiao": get_box_tiao_hedge_ratio(df_level[["Date"] + x_cols + [y_col]].set_index("Date"), dependent_variable=y_col),
        "johansen": get_johansen_hedge_ratio(df_level[["Date"] + x_cols + [y_col]].set_index("Date"), dependent_variable=y_col),
        "minimum_half_life": get_minimum_hl_hedge_ratio(df_level[["Date"] + x_cols + [y_col]].set_index("Date"), dependent_variable=y_col),
        "adf_optimal": get_adf_optimal_hedge_ratio(df_level[["Date"] + x_cols + [y_col]].set_index("Date"), dependent_variable=y_col),
        "pcr_pc1": (sm.OLS(df["spread"].to_numpy(), sm.add_constant(pc_scores_df["PC1"].to_numpy())).fit() if pc_scores_df is not None else None),
        "pcr_pc2": (sm.OLS(df["spread"].to_numpy(), sm.add_constant(pc_scores_df["PC2"].to_numpy())).fit() if pc_scores_df is not None else None),
        "pcr_pc3": (sm.OLS(df["spread"].to_numpy(), sm.add_constant(pc_scores_df["PC3"].to_numpy())).fit() if pc_scores_df is not None else None),
    }

    beta_estimates = {
        "ols": (
            regression_results["ols"].params[1] if len(x_cols) == 1 else [regression_results["ols"].params[1], regression_results["ols"].params[2]]
        ),
        "tls": (regression_results["tls"].beta[1] if len(x_cols) == 1 else [regression_results["tls"].beta[1], regression_results["tls"].beta[2]]),
        "odr": (
            regression_results["odr"].beta[1]
            if (x_errs is not None or y_errs is not None) and len(x_cols) == 1
            else [regression_results["odr"].beta[1], regression_results["odr"].beta[2]] if x_errs or y_errs else None
        ),
        "pc1": pc1_beta,
        "box_tiao": regression_results["box_tiao"]["beta_weights"],
        "johansen": regression_results["johansen"]["beta_weights"],
        "minimum_half_life": regression_results["minimum_half_life"]["beta_weights"],
        "adf_optimal": regression_results["adf_optimal"]["beta_weights"],
    }

    pcs_exposures = {
        "pcr_pc1_exposure": regression_results["pcr_pc1"].params[1] if pc_scores_df is not None else None,
        "pcr_pc2_exposure": regression_results["pcr_pc2"].params[1] if pc_scores_df is not None else None,
        "pcr_pc3_exposure": (regression_results["pcr_pc3"].params[1] if pc_scores_df is not None and len(x_cols) > 1 else None),
        # checking 50-50 duration - if non-zero => exposures exists
        "epsilon_pc1_loadings_exposure": (
            ep_y_pc1 - (ep_x0_pc1 + ep_x1_pc1) / 2.0 if len(x_cols) > 1 else ep_x0_pc1 - ep_y_pc1 if loadings_df is not None else None
        ),
        "epsilon_pc2_loadings_exposure": (
            ep_y_pc2 - (ep_x0_pc2 + ep_x1_pc2) / 2.0 if len(x_cols) > 1 else ep_x0_pc2 - ep_y_pc2 if loadings_df is not None else None
        ),
        "epsilon_pc3_loadings_exposure": (
            (ep_y_pc3 - (ep_x0_pc3 + ep_x1_pc3) / 2.0 if len(x_cols) > 1 else ep_x0_pc3 - ep_y_pc3 if loadings_df is not None else None)
            if len(x_cols) > 1
            else None
        ),
    }

    return {"betas": beta_estimates, "regression_results": regression_results, "pc_exposures": pcs_exposures}


# TODO
def rolling_beta_estimates():
    pass



import asyncio
import warnings
from datetime import datetime
from typing import Dict, List, Optional

import httpx
import pandas as pd
import requests

from CurvyCUSIPs.DataFetcher.base import DataFetcherBase
from CurvyCUSIPs.utils.ust_utils import build_treasurydirect_header

warnings.filterwarnings("ignore", category=pd.errors.SettingWithCopyWarning)
warnings.simplefilter(action="ignore", category=FutureWarning)

import sys

if sys.platform == "win32":
    loop = asyncio.ProactorEventLoop()
    asyncio.set_event_loop(loop)


class NYFRBDataFetcher(DataFetcherBase):
    def __init__(
        self,
        global_timeout: int = 10,
        proxies: Optional[Dict[str, str]] = None,
        debug_verbose: Optional[bool] = False,
        info_verbose: Optional[bool] = False,
        error_verbose: Optional[bool] = False,
    ):
        super().__init__(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

    async def _fetch_single_soma_holding_day(
        self,
        client: httpx.AsyncClient,
        date: datetime,
        valid_soma_dates_from_input: Dict[datetime, datetime],
        uid: Optional[str | int] = None,
        minimize_api_calls=False,
        max_retries: Optional[int] = 3,
        backoff_factor: Optional[int] = 1,
    ):
        cols_to_return = [
            "cusip",
            "asOfDate",
            "parValue",
            "percentOutstanding",
            "est_outstanding_amt",
            # "changeFromPriorWeek",
            # "changeFromPriorYear",
        ]
        retries = 0
        try:
            while retries < max_retries:
                try:
                    if minimize_api_calls:
                        date_str = date.strftime("%Y-%m-%d")
                    else:
                        date_str = valid_soma_dates_from_input[date].strftime("%Y-%m-%d")

                    url = f"https://markets.newyorkfed.org/api/soma/tsy/get/asof/{date_str}.json"
                    response = await client.get(
                        url,
                        headers=build_treasurydirect_header(host_str="markets.newyorkfed.org"),
                    )
                    response.raise_for_status()
                    curr_soma_holdings_json = response.json()
                    curr_soma_holdings_df = pd.DataFrame(curr_soma_holdings_json["soma"]["holdings"])
                    curr_soma_holdings_df = curr_soma_holdings_df.fillna("")
                    curr_soma_holdings_df["asOfDate"] = pd.to_datetime(curr_soma_holdings_df["asOfDate"], errors="coerce")
                    curr_soma_holdings_df["parValue"] = pd.to_numeric(curr_soma_holdings_df["parValue"], errors="coerce")
                    curr_soma_holdings_df["percentOutstanding"] = pd.to_numeric(curr_soma_holdings_df["percentOutstanding"], errors="coerce")
                    curr_soma_holdings_df["est_outstanding_amt"] = curr_soma_holdings_df["parValue"] / curr_soma_holdings_df["percentOutstanding"]
                    curr_soma_holdings_df = curr_soma_holdings_df[
                        (curr_soma_holdings_df["securityType"] != "TIPS") & (curr_soma_holdings_df["securityType"] != "FRNs")
                    ]
                    if uid:
                        return date, curr_soma_holdings_df[cols_to_return], uid

                    return date, curr_soma_holdings_df[cols_to_return]

                except httpx.HTTPStatusError as e:
                    self._logger.error(f"SOMA Holding - Bad Status: {response.status_code}")
                    if response.status_code == 404:
                        if uid:
                            return date, pd.DataFrame(columns=cols_to_return), uid
                        return date, pd.DataFrame(columns=cols_to_return)

                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"SOMA Holding - Throttled. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

                except Exception as e:
                    self._logger.error(f"SOMA Holding - Error: {str(e)}")
                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"SOMA Holding - Throttled. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

            raise ValueError(f"UST STRIPPING Activity - Max retries exceeded for {date}")

        except Exception as e:
            self._logger.error(e)
            if uid:
                return date, pd.DataFrame(columns=cols_to_return), uid
            return date, pd.DataFrame(columns=cols_to_return)

    async def _fetch_single_soma_holding_day_with_semaphore(self, semaphore, *args, **kwargs):
        async with semaphore:
            return await self._fetch_single_soma_holding_day(*args, **kwargs)

    async def _build_fetch_tasks_historical_soma_holdings(
        self,
        client: httpx.AsyncClient,
        dates: List[datetime],
        minimize_api_calls: Optional[bool] = False,
        uid: Optional[str | int] = None,
        max_retries: Optional[int] = 3,
        backoff_factor: Optional[int] = 1,
        max_concurrent_tasks: int = 64,
        my_semaphore: Optional[asyncio.Semaphore] = None,
    ):
        valid_soma_holding_dates_reponse = requests.get(
            "https://markets.newyorkfed.org/api/soma/asofdates/list.json",
            headers=build_treasurydirect_header(host_str="markets.newyorkfed.org"),
            proxies=self._proxies,
        )
        if valid_soma_holding_dates_reponse.ok:
            valid_soma_holding_dates_json = valid_soma_holding_dates_reponse.json()
            valid_soma_dates_dt = [datetime.strptime(dt_string, "%Y-%m-%d") for dt_string in valid_soma_holding_dates_json["soma"]["asOfDates"]]
        else:
            raise ValueError(f"SOMA Holdings - Status Code: {valid_soma_holding_dates_reponse.status_code}")

        valid_soma_dates_from_input = {}
        for dt in dates:
            valid_closest_date = min(
                (valid_date for valid_date in valid_soma_dates_dt if valid_date <= dt),
                key=lambda valid_date: abs(dt - valid_date),
            )
            valid_soma_dates_from_input[dt] = valid_closest_date
        self._logger.debug(f"SOMA Holdings - Valid SOMA Holding Dates: {valid_soma_dates_from_input}")

        semaphore = my_semaphore or asyncio.Semaphore(max_concurrent_tasks)
        if minimize_api_calls:
            tasks = [
                self._fetch_single_soma_holding_day_with_semaphore(
                    semaphore=semaphore,
                    client=client,
                    date=date,
                    valid_soma_dates_from_input=valid_soma_dates_from_input,
                    uid=uid,
                    minimize_api_calls=True,
                    max_retries=max_retries,
                    backoff_factor=backoff_factor,
                )
                for date in list(set(valid_soma_dates_from_input.values()))
            ]
            return tasks
        else:
            tasks = [
                self._fetch_single_soma_holding_day_with_semaphore(
                    semaphore=semaphore,
                    client=client,
                    date=date,
                    valid_soma_dates_from_input=valid_soma_dates_from_input,
                    uid=uid,
                    max_retries=max_retries,
                    backoff_factor=backoff_factor,
                )
                for date in dates
            ]
            return tasks



# stolen from https://github.com/luphord/nelson_siegel_svensson/blob/master/nelson_siegel_svensson/nss.py
# -*- coding: utf-8 -*-

"""Implementation of a Nelson-Siegel-Svensson interest rate curve model.
See `NelsonSiegelSvenssonCurve` class for details.
"""

from numbers import Real
from dataclasses import dataclass
from typing import Union, Tuple

import numpy as np
from numpy import exp

EPS = np.finfo(float).eps


@dataclass
class NelsonSiegelSvenssonCurve:
    """Implementation of a Nelson-Siegel-Svensson interest rate curve model.
    This curve can be interpreted as a factor model with four
    factors (including a constant).
    """

    beta0: float
    beta1: float
    beta2: float
    beta3: float
    tau1: float
    tau2: float

    def factors(
        self, T: Union[float, np.ndarray]
    ) -> Union[Tuple[float, float, float], Tuple[np.ndarray, np.ndarray, np.ndarray]]:
        """Factor loadings for time(s) T, excluding constant."""
        tau1 = self.tau1
        tau2 = self.tau2
        if isinstance(T, Real) and T <= 0:
            return 1, 0, 0
        elif isinstance(T, np.ndarray):
            zero_idx = T <= 0
            T[zero_idx] = EPS  # avoid warnings in calculations
        exp_tt1 = exp(-T / tau1)
        exp_tt2 = exp(-T / tau2)
        factor1 = (1 - exp_tt1) / (T / tau1)
        factor2 = factor1 - exp_tt1
        factor3 = (1 - exp_tt2) / (T / tau2) - exp_tt2
        if isinstance(T, np.ndarray):
            T[zero_idx] = 0
            factor1[zero_idx] = 1
            factor2[zero_idx] = 0
            factor3[zero_idx] = 0
        return factor1, factor2, factor3

    def factor_matrix(self, T: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Factor loadings for time(s) T as matrix columns,
        including constant column (=1.0).
        """
        factor1, factor2, factor3 = self.factors(T)
        constant: Union[float, np.ndarray] = (
            np.ones(T.size) if isinstance(T, np.ndarray) else 1.0
        )
        return np.stack([constant, factor1, factor2, factor3]).transpose()

    def zero(self, T: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Zero rate(s) of this curve at time(s) T."""
        beta0 = self.beta0
        beta1 = self.beta1
        beta2 = self.beta2
        beta3 = self.beta3
        factor1, factor2, factor3 = self.factors(T)
        res = beta0 + beta1 * factor1 + beta2 * factor2 + beta3 * factor3
        return res

    def __call__(self, T: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Zero rate(s) of this curve at time(s) T."""
        return self.zero(T)

    def forward(self, T: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Instantaneous forward rate(s) of this curve at time(s) T."""
        exp_tt0 = exp(-T / self.tau1)
        exp_tt1 = exp(-T / self.tau2)
        return (
            self.beta0
            + self.beta1 * exp_tt0
            + self.beta2 * exp_tt0 * T / self.tau1
            + self.beta3 * exp_tt1 * T / self.tau2
        )



import warnings
from concurrent.futures import ProcessPoolExecutor, as_completed
from datetime import datetime, timedelta
from typing import Dict, List, Literal, Optional

import numpy as np
import pandas as pd
import QuantLib as ql
import rateslib as rl
import scipy.interpolate
from pandas.tseries.offsets import BDay

from CurvyCUSIPs.CurveDataFetcher import CurveDataFetcher
from CurvyCUSIPs.utils.ust_utils import pydatetime_to_quantlib_date, quantlib_date_to_pydatetime, to_quantlib_fixed_rate_bond_obj

warnings.filterwarnings("ignore", category=pd.errors.SettingWithCopyWarning)
warnings.simplefilter(action="ignore", category=FutureWarning)


def _calc_spot_rates_on_tenors(
    yield_curve: ql.DiscountCurve | ql.ZeroCurve,
    on_rate: float,
    day_count: ql.ActualActual = ql.ActualActual(ql.ActualActual.ISDA),
    price_col: Optional[str] = None,
    custom_price_col: Optional[str] = None,
    continuous_compounded_zero: Optional[bool] = False,
):
    spots = []
    tenors = []
    maturity_dates = []
    ref_date = yield_curve.referenceDate()

    dates = yield_curve.dates()
    for i, d in enumerate(dates):
        yrs = day_count.yearFraction(ref_date, d)
        if i == 0:
            tenors.append(1 / 360)
            spots.append(on_rate)
            t_plus_1_sr: pd.Timestamp = quantlib_date_to_pydatetime(d) - BDay(1)
            t_plus_1_sr = t_plus_1_sr.to_pydatetime()
            t_plus_1_sr = t_plus_1_sr.replace(hour=0, minute=0, second=0, microsecond=0)
            maturity_dates.append(t_plus_1_sr)
            continue

        if continuous_compounded_zero:
            zero_rate = yield_curve.zeroRate(yrs, ql.Continuous, True)
            eq_rate = zero_rate.equivalentRate(day_count, ql.Continuous, ql.NoFrequency, ref_date, d).rate()
        else:
            compounding = ql.Compounded
            freq = ql.Semiannual
            zero_rate = yield_curve.zeroRate(yrs, compounding, freq, True)
            eq_rate = zero_rate.equivalentRate(day_count, compounding, freq, ref_date, d).rate()

        tenors.append(yrs)
        spots.append(100 * eq_rate)
        maturity_dates.append(quantlib_date_to_pydatetime(d))

    price_col_type = price_col.split("_")[0] if price_col else None
    spot_col_name = f"{price_col_type}_spot_rate" if price_col else "spot_rate"
    if custom_price_col:
        spot_col_name = custom_price_col
    return pd.DataFrame(
        {
            "maturity_date": maturity_dates,
            "time_to_maturity": tenors,
            spot_col_name: spots,
        }
    )


def _calc_spot_rates_intep_months(
    yield_curve: ql.DiscountCurve | ql.ZeroCurve,
    on_rate: float,
    months: Optional[int] = 361,
    month_freq: Optional[float] = 1,
    custom_tenors: Optional[List[int]] = None,
    day_count=ql.ActualActual(ql.ActualActual.ISDA),
    calendar=ql.UnitedStates(m=ql.UnitedStates.GovernmentBond),
    price_col: Optional[str] = None,
    custom_price_col: Optional[str] = None,
    continuous_compounded_zero: Optional[bool] = False,
):
    spots = []
    tenors = []
    maturity_dates = []
    ref_date = yield_curve.referenceDate()
    to_iterate = custom_tenors if custom_tenors else range(0, months, month_freq)
    for month in to_iterate:
        d = calendar.advance(ref_date, ql.Period(month, ql.Months))
        yrs = month / 12.0
        if yrs == 0:
            tenors.append(1 / 360)
            spots.append(on_rate)
            t_plus_1_sr: pd.Timestamp = quantlib_date_to_pydatetime(d) - BDay(1)
            t_plus_1_sr = t_plus_1_sr.to_pydatetime()
            t_plus_1_sr = t_plus_1_sr.replace(hour=0, minute=0, second=0, microsecond=0)
            maturity_dates.append(t_plus_1_sr)
            continue

        if continuous_compounded_zero:
            zero_rate = yield_curve.zeroRate(yrs, ql.Continuous, True)
            eq_rate = zero_rate.equivalentRate(day_count, ql.Continuous, ql.NoFrequency, ref_date, d).rate()
        else:
            compounding = ql.Compounded
            freq = ql.Semiannual
            zero_rate = yield_curve.zeroRate(yrs, compounding, freq, True)
            eq_rate = zero_rate.equivalentRate(day_count, compounding, freq, ref_date, d).rate()

        tenors.append(yrs)
        spots.append(100 * eq_rate)
        maturity_dates.append(quantlib_date_to_pydatetime(d))

    price_col_type = price_col.split("_")[0] if price_col else None
    spot_col_name = f"{price_col_type}_spot_rate" if price_col else "spot_rate"
    if custom_price_col:
        spot_col_name = custom_price_col
    return pd.DataFrame(
        {
            "maturity_date": maturity_dates,
            "time_to_maturity": tenors,
            spot_col_name: spots,
        }
    )


def _calc_spot_rates_intep_days(
    yield_curve: ql.DiscountCurve | ql.ZeroCurve,
    on_rate: float,
    days: Optional[int] = 361 * 30,
    custom_tenors: Optional[List[int]] = None,
    day_count=ql.ActualActual(ql.ActualActual.ISDA),
    calendar=ql.UnitedStates(m=ql.UnitedStates.GovernmentBond),
    price_col: Optional[str] = None,
    custom_price_col: Optional[str] = None,
    continuous_compounded_zero: Optional[bool] = False,
):
    spots = []
    tenors = []
    maturity_dates = []
    ref_date = yield_curve.referenceDate()
    to_iterate = custom_tenors if custom_tenors else range(0, days, 1)
    for day in to_iterate:
        d = calendar.advance(ref_date, ql.Period(day, ql.Days))
        yrs = day / 365.0
        if yrs == 0:
            tenors.append(1 / 360)
            spots.append(on_rate)
            t_plus_1_sr: pd.Timestamp = quantlib_date_to_pydatetime(d) - BDay(1)
            t_plus_1_sr = t_plus_1_sr.to_pydatetime()
            t_plus_1_sr = t_plus_1_sr.replace(hour=0, minute=0, second=0, microsecond=0)
            maturity_dates.append(t_plus_1_sr)
            continue

        if continuous_compounded_zero:
            zero_rate = yield_curve.zeroRate(yrs, ql.Continuous, True)
            eq_rate = zero_rate.equivalentRate(day_count, ql.Continuous, ql.NoFrequency, ref_date, d).rate()
        else:
            compounding = ql.Compounded
            freq = ql.Semiannual
            zero_rate = yield_curve.zeroRate(yrs, compounding, freq, True)
            eq_rate = zero_rate.equivalentRate(day_count, compounding, freq, ref_date, d).rate()

        tenors.append(yrs)
        spots.append(100 * eq_rate)
        maturity_dates.append(quantlib_date_to_pydatetime(d))

    price_col_type = price_col.split("_")[0] if price_col else None
    spot_col_name = f"{price_col_type}_spot_rate" if price_col else "spot_rate"
    if custom_price_col:
        spot_col_name = custom_price_col
    return pd.DataFrame(
        {
            "maturity_date": maturity_dates,
            "time_to_maturity": tenors,
            spot_col_name: spots,
        }
    )


"""
Using QuantLib's Piecewise yield term structure for bootstrapping market observed prices to zeros rates at the respective ttms
- small differences between methods
- flag to take the averages of all Piecewise methods or pass in a specifc method
- passing in multiple ql_bootstrap_methods will take the average of the spot rates calculated from the different methods 
"""


def ql_piecewise_method_pretty(bs_method):
    ql_piecewise_methods_pretty_dict = {
        "ql_plld": "Piecewise Log Linear Discount",
        "ql_lcd": "Piecewise Log Cubic Discount",
        "ql_lz": "Piecewise Linear Zero",
        "ql_cz": "Piecewise Cubic Zero",
        "ql_lf": "Piecewise Linear Forward",
        "ql_spd": "Piecewise Spline Cubic Discount",
        "ql_kz": "Piecewise Kruger Zero",
        "ql_kld": "Piecewise Kruger Log Discount",
        "ql_mcf": "Piecewise Convex Monotone Forward",
        "ql_mcz": "Piecewise Convex Monotone Zero",
        "ql_ncz": "Piecewise Natural Cubic Zero",
        "ql_nlcd": "Piecewise Natural Log Cubic Discount",
        "ql_lmlcd": "Piecewise Log Mixed Linear Cubic Discount",
        "ql_pcz": "Piecewise Parabolic Cubic Zero",
        "ql_mpcz": "Piecewise Monotonic Parabolic Cubic Zero",
        "ql_lpcd": "Piecewise Log Parabolic Cubic Discount",
        "ql_mlpcd": "Piecewise Monotonic Log Parabolic Cubic Discount",
        "ql_f_ns": "Nelson-Siegel Fitting",
        "ql_f_nss": "Svensson Fitting",
        "ql_f_np": "Simple Polynomial Fitting",
        "ql_f_es": "Exponential Splines Fitting",
        "ql_f_cbs": "Cubic B-Splines Fitting",
    }
    return ql_piecewise_methods_pretty_dict[bs_method]


def get_spot_rates_bootstrapper(
    curve_set_df: pd.DataFrame,
    as_of_date: datetime,
    on_rate: float,
    ql_bootstrap_interp_methods: Optional[
        List[
            Literal[
                "ql_plld",
                "ql_lcd",
                "ql_lz",
                "ql_cz",
                "ql_lf",
                "ql_spd",
                "ql_kz",
                "ql_kld",
                "ql_mcf",
                "ql_mcz",
                "ql_ncz",
                "ql_nlcd",
                "ql_lmlcd",
                "ql_pcz",
                "ql_mpcz",
                "ql_lpcd",
                "ql_mlpcd",
            ]
        ]
    ] = ["ql_plld"],
    return_ql_zero_curve: Optional[bool] = False,
    interpolated_months_num: Optional[int] = None,
    interpolated_curve_yearly_freq: Optional[int] = 1,
    custom_yearly_tenors: Optional[List[int]] = None,
    # return_rel_val_df: Optional[bool] = False,
    daily_interpolation: Optional[bool] = False,
    return_scipy_interp_func: Optional[bool] = False,
    continuous_compounded_zero: Optional[bool] = False,
) -> Dict[str, pd.DataFrame | ql.DiscountCurve | ql.ZeroCurve]:
    ql_piecewise_methods: Dict[str, ql.DiscountCurve | ql.ZeroCurve] = {
        "ql_plld": ql.PiecewiseLogLinearDiscount,
        "ql_lcd": ql.PiecewiseLogCubicDiscount,
        "ql_lz": ql.PiecewiseLinearZero,
        "ql_cz": ql.PiecewiseCubicZero,
        "ql_lf": ql.PiecewiseLinearForward,
        "ql_spd": ql.PiecewiseSplineCubicDiscount,
        "ql_kz": ql.PiecewiseKrugerZero,
        "ql_kld": ql.PiecewiseKrugerLogDiscount,
        "ql_mcf": ql.PiecewiseConvexMonotoneForward,
        "ql_mcz": ql.PiecewiseConvexMonotoneZero,
        "ql_ncz": ql.PiecewiseNaturalCubicZero,
        "ql_nlcd": ql.PiecewiseNaturalLogCubicDiscount,
        "ql_lmlcd": ql.PiecewiseLogMixedLinearCubicDiscount,
        "ql_pcz": ql.PiecewiseParabolicCubicZero,
        "ql_mpcz": ql.PiecewiseMonotonicParabolicCubicZero,
        "ql_lpcd": ql.PiecewiseLogParabolicCubicDiscount,
        "ql_mlpcd": ql.PiecewiseMonotonicLogParabolicCubicDiscount,
        "ql_f_ns": ql.NelsonSiegelFitting,
        "ql_f_nss": ql.SvenssonFitting,
        "ql_f_np": ql.SimplePolynomialFitting,
        "ql_f_es": ql.ExponentialSplinesFitting,
        "ql_f_cbs": ql.CubicBSplinesFitting,
    }

    price_cols = ["bid_price", "offer_price", "mid_price", "eod_price"]
    required_cols = ["issue_date", "maturity_date", "int_rate"]
    price_col_exists = any(item in curve_set_df.columns for item in price_cols)
    missing_required_cols = [item for item in required_cols if item not in curve_set_df.columns]

    if not price_col_exists:
        raise ValueError(f"Build Spot Curve - Couldn't find a valid price col in your curve set df - one of {price_cols}")
    if missing_required_cols:
        raise ValueError(f"Build Spot Curve - Missing required curve set cols: {missing_required_cols}")

    price_col = next((item for item in price_cols if item in curve_set_df.columns), None)
    calendar = ql.UnitedStates(m=ql.UnitedStates.GovernmentBond)
    today = calendar.adjust(pydatetime_to_quantlib_date(py_datetime=as_of_date))
    ql.Settings.instance().evaluationDate = today

    t_plus = 1
    bond_settlement_date = calendar.advance(today, ql.Period(t_plus, ql.Days))
    frequency = ql.Semiannual
    day_count = ql.ActualActual(ql.ActualActual.ISDA)
    par = 100.0

    bond_helpers = []
    for _, row in curve_set_df.iterrows():
        maturity = pydatetime_to_quantlib_date(row["maturity_date"])
        if np.isnan(row["int_rate"]):
            quote = ql.QuoteHandle(ql.SimpleQuote(row[price_col]))
            tbill = ql.ZeroCouponBond(
                t_plus,
                calendar,
                par,
                maturity,
                ql.ModifiedFollowing,
                100.0,
                bond_settlement_date,
            )
            helper = ql.BondHelper(quote, tbill)
        else:
            schedule = ql.Schedule(
                bond_settlement_date,
                maturity,
                ql.Period(frequency),
                calendar,
                ql.ModifiedFollowing,
                ql.ModifiedFollowing,
                ql.DateGeneration.Backward,
                False,
            )
            helper = ql.FixedRateBondHelper(
                ql.QuoteHandle(ql.SimpleQuote(row[price_col])),
                t_plus,
                100.0,
                schedule,
                [row["int_rate"] / 100],
                day_count,
                ql.ModifiedFollowing,
                par,
            )

        bond_helpers.append(helper)

    spot_dfs: List[pd.DataFrame] = []
    ql_curves: Dict[str, ql.DiscountCurve | ql.ZeroCurve] = {}

    for bs_method in ql_bootstrap_interp_methods:
        if bs_method.split("_")[1] == "f":
            ql_fit_method = ql_piecewise_methods[bs_method]
            curr_curve = ql.FittedBondDiscountCurve(bond_settlement_date, bond_helpers, day_count, ql_fit_method())
            curr_curve.enableExtrapolation()
            ql_curves[bs_method] = curr_curve
        else:
            curr_curve = ql_piecewise_methods[bs_method](bond_settlement_date, bond_helpers, day_count)
            curr_curve.enableExtrapolation()
            ql_curves[bs_method] = curr_curve
        if interpolated_months_num or custom_yearly_tenors:
            if daily_interpolation:
                curr_spot_df = _calc_spot_rates_intep_days(
                    yield_curve=curr_curve,
                    on_rate=on_rate,
                    days=interpolated_months_num * 31,
                    custom_price_col=f"{bs_method}_spot_rate",
                    continuous_compounded_zero=continuous_compounded_zero,
                )
            else:
                curr_spot_df = _calc_spot_rates_intep_months(
                    yield_curve=curr_curve,
                    on_rate=on_rate,
                    months=interpolated_months_num,
                    month_freq=interpolated_curve_yearly_freq,
                    custom_tenors=([i * 12 for i in custom_yearly_tenors] if custom_yearly_tenors else None),
                    custom_price_col=f"{bs_method}_spot_rate",
                    continuous_compounded_zero=continuous_compounded_zero,
                )
        else:
            curr_spot_df = _calc_spot_rates_on_tenors(
                yield_curve=curr_curve,
                on_rate=on_rate,
                custom_price_col=f"{bs_method}_spot_rate",
                continuous_compounded_zero=continuous_compounded_zero,
            )

        spot_dfs.append(curr_spot_df)

    if len(spot_dfs) == 1:
        zero_rates_df = spot_dfs[0]
    else:
        maturity_dates = spot_dfs[0]["maturity_date"].to_list()
        tenors = spot_dfs[0]["time_to_maturity"].to_list()
        merged_df = pd.concat(
            [df[[col for col in df.columns if "spot_rate" in col]] for df in spot_dfs],
            axis=1,
        )
        avg_spot_rate_col = merged_df.mean(axis=1).to_list()

        merged_df.insert(0, "maturity_date", maturity_dates)
        merged_df.insert(1, "time_to_maturity", tenors)
        merged_df["avg_spot_rate"] = avg_spot_rate_col
        zero_rates_df = merged_df

    to_return_dict = {
        "ql_zero_curve_obj": None,
        "spot_rate_df": None,
        "scipy_interp_funcs": None,
    }

    if return_ql_zero_curve:
        if len(ql_bootstrap_interp_methods) > 1:
            print("Get Spot Rates - multiple bs methods passed - returning ql zero curve based on first bs method")
        bs_method = ql_bootstrap_interp_methods[0]
        if bs_method.split("_")[1] == "f":
            ql_fit_method = ql_piecewise_methods[bs_method]
            zero_curve = ql.FittedBondDiscountCurve(bond_settlement_date, bond_helpers, day_count, ql_fit_method())
        else:
            zero_curve = ql_piecewise_methods[bs_method](bond_settlement_date, bond_helpers, day_count)
        zero_curve.enableExtrapolation()
        to_return_dict["ql_zero_curve_obj"] = zero_curve

    to_return_dict["spot_rate_df"] = zero_rates_df

    if return_scipy_interp_func:
        scipy_interp_funcs = {}
        for bs_method in ql_bootstrap_interp_methods:
            scipy_interp_funcs[bs_method] = scipy.interpolate.interp1d(
                to_return_dict["spot_rate_df"]["time_to_maturity"],
                to_return_dict["spot_rate_df"][f"{bs_method}_spot_rate"],
                axis=0,
                kind="linear",
                # bounds_error=False,
                # fill_value="extrapolate",
            )
        to_return_dict["scipy_interp_funcs"] = scipy_interp_funcs

    return to_return_dict


def get_par_rates(
    spot_rates: List[float],
    tenors: List[int],
    select_every_nth_spot_rate: Optional[int] = None,
) -> pd.DataFrame:
    if select_every_nth_spot_rate:
        spot_rates = spot_rates[::select_every_nth_spot_rate]
    par_rates = []
    for tenor in tenors:
        periods = np.arange(0, tenor + 0.5, 0.5)
        curr_spot_rates = spot_rates[: len(periods)].copy()
        discount_factors = [1 / (1 + (s / 100) / 2) ** (2 * t) for s, t in zip(curr_spot_rates, periods)]
        sum_of_dfs = sum(discount_factors[:-1])
        par_rate = (1 - discount_factors[-1]) / sum_of_dfs * 2
        par_rates.append(par_rate * 100)

    return pd.DataFrame(
        {
            "tenor": tenors,
            "par_rate": par_rates,
        }
    )


# TODO match ql interp method with scipy interp func
def get_spot_rates_fitter(
    curve_set_df: pd.DataFrame,
    as_of_date: datetime,
    on_rate: float,
    ql_fitting_methods: Optional[
        List[
            Literal[
                "ql_f_ns",
                "ql_f_nss",
                "ql_f_sp",
                "ql_f_es",
                "ql_f_cbs",
            ]
        ]
    ] = ["ql_f_nss"],
    ql_zero_curve_interp_method: Optional[
        Literal[
            "ql_z_interp_log_lin",
            "ql_z_interp_cubic",
            "ql_z_interp_nat_cubic",
            "ql_z_interp_log_cubic",
            "ql_z_interp_monot_cubic",
        ]
    ] = None,
    daily_interpolation: Optional[bool] = False,
    simple_poly: Optional[int] = None,
    knots: Optional[List[float]] = None,
):
    ql_fitting_methods_dict: Dict[str, ql.DiscountCurve | ql.ZeroCurve] = {
        "ql_f_ns": ql.NelsonSiegelFitting,
        "ql_f_nss": ql.SvenssonFitting,
        "ql_f_sp": ql.SimplePolynomialFitting,
        "ql_f_es": ql.ExponentialSplinesFitting,
        "ql_f_cbs": ql.CubicBSplinesFitting,
    }

    ql_zero_curve_interp_methods_dict: Dict[str, ql.ZeroCurve] = {
        "ql_z_interp_log_lin": ql.LogLinearZeroCurve,
        "ql_z_interp_cubic": ql.CubicZeroCurve,
        "ql_z_interp_nat_cubic": ql.NaturalCubicZeroCurve,
        "ql_z_interp_log_cubic": ql.LogCubicZeroCurve,
        "ql_z_interp_monot_cubic": ql.MonotonicCubicZeroCurve,
    }

    price_cols = ["bid_price", "offer_price", "mid_price", "eod_price"]
    required_cols = ["issue_date", "maturity_date", "int_rate"]
    price_col_exists = any(item in curve_set_df.columns for item in price_cols)
    missing_required_cols = [item for item in required_cols if item not in curve_set_df.columns]

    if not price_col_exists:
        raise ValueError(f"Build Spot Curve - Couldn't find a valid price col in your curve set df - one of {price_cols}")
    if missing_required_cols:
        raise ValueError(f"Build Spot Curve - Missing required curve set cols: {missing_required_cols}")

    price_col = next((item for item in price_cols if item in curve_set_df.columns), None)
    calendar = ql.UnitedStates(m=ql.UnitedStates.GovernmentBond)
    today = calendar.adjust(pydatetime_to_quantlib_date(py_datetime=as_of_date))
    ql.Settings.instance().evaluationDate = today

    t_plus = 1
    bond_settlement_date = calendar.advance(today, ql.Period(t_plus, ql.Days))
    frequency = ql.Semiannual
    day_count = ql.ActualActual(ql.ActualActual.ISDA)
    par = 100.0

    bond_helpers = []
    for _, row in curve_set_df.iterrows():
        maturity = pydatetime_to_quantlib_date(row["maturity_date"])
        if np.isnan(row["int_rate"]):
            quote = ql.QuoteHandle(ql.SimpleQuote(row[price_col]))
            tbill = ql.ZeroCouponBond(
                t_plus,
                calendar,
                par,
                maturity,
                ql.ModifiedFollowing,
                100.0,
                bond_settlement_date,
            )
            helper = ql.BondHelper(quote, tbill)
        else:
            schedule = ql.Schedule(
                bond_settlement_date,
                maturity,
                ql.Period(frequency),
                calendar,
                ql.ModifiedFollowing,
                ql.ModifiedFollowing,
                ql.DateGeneration.Backward,
                False,
            )
            helper = ql.FixedRateBondHelper(
                ql.QuoteHandle(ql.SimpleQuote(row[price_col])),
                t_plus,
                100.0,
                schedule,
                [row["int_rate"] / 100],
                day_count,
                ql.ModifiedFollowing,
                par,
            )

        bond_helpers.append(helper)

    ql_curves: Dict[str, Dict[str, ql.DiscountCurve | ql.ZeroCurve | pd.DataFrame]] = {}
    for fit_method in ql_fitting_methods:
        if fit_method == "ql_f_sp" and not simple_poly:
            continue
        if fit_method == "ql_f_cbs" and not knots:
            continue

        ql_fit_method = ql_fitting_methods_dict[fit_method]

        if fit_method == "ql_f_sp":
            called_ql_fit_method = ql_fit_method(simple_poly)
        elif fit_method == "ql_f_cbs":
            called_ql_fit_method = ql_fit_method(knots)
        else:
            called_ql_fit_method = ql_fit_method()

        curr_curve = ql.FittedBondDiscountCurve(bond_settlement_date, bond_helpers, day_count, called_ql_fit_method)
        curr_curve.enableExtrapolation()
        if fit_method not in ql_curves:
            ql_curves[fit_method] = {
                "ql_fitted_curve": None,
                "ql_zero_curve": None,
                "zero_interp_func": None,
                "df_interp_func": None,
                "comparison_df": None,
            }

        ql_curves[fit_method]["ql_curve"] = curr_curve

        if daily_interpolation:
            dates = [bond_settlement_date + ql.Period(i, ql.Days) for i in range(0, 12 * 30 * 30, 1)]
        else:
            dates = [bond_settlement_date + ql.Period(i, ql.Months) for i in range(0, 12 * 30, 1)]

        discount_factors = [curr_curve.discount(d) for d in dates]
        ttm = [(ql.Date.to_date(d) - ql.Date.to_date(bond_settlement_date)).days / 365 for d in dates]

        eq_zero_rates = []
        eq_zero_rates_dec = []
        for d in dates:
            yrs = (ql.Date.to_date(d) - ql.Date.to_date(bond_settlement_date)).days / 365.0
            zero_rate = curr_curve.zeroRate(yrs, ql.Continuous, True)
            eq_rate = zero_rate.equivalentRate(day_count, ql.Continuous, ql.NoFrequency, bond_settlement_date, d).rate()
            eq_zero_rates.append(eq_rate * 100)
            eq_zero_rates_dec.append(eq_rate)
        eq_zero_rates[0] = on_rate
        eq_zero_rates_dec[0] = on_rate / 100

        ql_curves[fit_method]["zero_interp_func"] = scipy.interpolate.interp1d(ttm, eq_zero_rates, axis=0, kind="linear", fill_value="extrapolate")
        ql_curves[fit_method]["df_interp_func"] = scipy.interpolate.interp1d(ttm, discount_factors, axis=0, kind="linear", fill_value="extrapolate")
        zero_curve = (
            ql.ZeroCurve(dates, eq_zero_rates_dec, day_count)
            if not ql_zero_curve_interp_method
            else ql_zero_curve_interp_methods_dict[ql_zero_curve_interp_method](dates, eq_zero_rates_dec, day_count)
        )
        zero_curve.enableExtrapolation()
        ql_curves[fit_method]["ql_zero_curve"] = zero_curve

    return ql_curves


def reprice_bonds_single_zero_curve(as_of_date: datetime, ql_zero_curve: ql.ZeroCurve, curve_set_df: pd.DataFrame) -> pd.DataFrame:
    yield_curve_handle = ql.YieldTermStructureHandle(ql_zero_curve)
    engine = ql.DiscountingBondEngine(yield_curve_handle)

    price_cols = ["bid_price", "offer_price", "mid_price", "eod_price"]
    required_cols = ["issue_date", "maturity_date", "int_rate"]
    price_col_exists = any(item in curve_set_df.columns for item in price_cols)
    missing_required_cols = [item for item in required_cols if item not in curve_set_df.columns]

    if not price_col_exists:
        raise ValueError(f"Build Spot Curve - Couldn't find a valid price col in your curve set df - one of {price_cols}")
    if missing_required_cols:
        raise ValueError(f"Build Spot Curve - Missing required curve set cols: {missing_required_cols}")

    price_col = next((item for item in price_cols if item in curve_set_df.columns), None)
    quote_type = price_col.split("_")[0]
    yield_col = f"{quote_type}_yield"

    if yield_col not in curve_set_df.columns:
        raise ValueError(f"Build Spot Curve - Missing required curve set cols: {yield_col}")

    calendar = ql.UnitedStates(m=ql.UnitedStates.GovernmentBond)
    today = calendar.adjust(pydatetime_to_quantlib_date(py_datetime=as_of_date))
    ql.Settings.instance().evaluationDate = today

    t_plus = 1
    bond_settlement_date = calendar.advance(today, ql.Period(t_plus, ql.Days))
    frequency = ql.Semiannual
    day_count = ql.ActualActual(ql.ActualActual.ISDA)
    par = 100.0

    bonds: List[Dict[str, ql.FixedRateBond | ql.ZeroCouponBond]] = []
    for _, row in curve_set_df.iterrows():
        try:
            maturity = pydatetime_to_quantlib_date(row["maturity_date"])
            if np.isnan(row["int_rate"]):
                bond_ql = ql.ZeroCouponBond(
                    t_plus,
                    calendar,
                    par,
                    maturity,
                    ql.ModifiedFollowing,
                    100.0,
                    bond_settlement_date,
                )
                bond_rl: rl.FixedRateBond = rl.Bill(
                    termination=row["maturity_date"],
                    effective=row["issue_date"],
                    calendar="nyc",
                    modifier="NONE",
                    currency="usd",
                    convention="Act360",
                    settle=1,
                    curves="bill_curve",
                    calc_mode="us_gbb",
                )
            else:
                schedule = ql.Schedule(
                    bond_settlement_date,
                    maturity,
                    ql.Period(frequency),
                    calendar,
                    ql.ModifiedFollowing,
                    ql.ModifiedFollowing,
                    ql.DateGeneration.Backward,
                    False,
                )
                bond_ql = ql.FixedRateBond(
                    t_plus,
                    100.0,
                    schedule,
                    [row["int_rate"] / 100],
                    day_count,
                    ql.ModifiedFollowing,
                )
                bond_rl: rl.FixedRateBond = rl.FixedRateBond(
                    effective=row["issue_date"],
                    termination=row["maturity_date"],
                    fixed_rate=row["int_rate"],
                    spec="ust",
                    calc_mode="ust_31bii",
                )

            curr_accrued_amount = bond_rl.accrued(quantlib_date_to_pydatetime(bond_settlement_date))
            bond_ql.setPricingEngine(engine)
            curr_calced_npv = bond_ql.NPV()
            curr_calced_ytm = (
                bond_ql.bondYield(
                    curr_calced_npv,
                    day_count,
                    ql.Compounded,
                    frequency,
                    bond_settlement_date,
                )
                * 100
            )
            bonds.append(
                {
                    "cusip": row["cusip"],
                    "label": row["label"],
                    "issue_date": row["issue_date"],
                    "maturity_date": row["maturity_date"],
                    "time_to_maturity": row["time_to_maturity"],
                    "high_investment_rate": row["high_investment_rate"],
                    "int_rate": row["int_rate"],
                    "rank": row["rank"] if "rank" in curve_set_df.columns else None,
                    "outstanding": (row["outstanding_amt"] if "outstanding_amt" in curve_set_df.columns else None),
                    "soma_holdings": (row["parValue"] if "parValue" in curve_set_df.columns else None),
                    "stripping_amount": (row["portion_stripped_amt"] if "portion_stripped_amt" in curve_set_df.columns else None),
                    "free_float": (row["free_float"] if "free_float" in curve_set_df.columns else None),
                    yield_col: row[yield_col],
                    price_col: row[price_col],
                    "accured": curr_accrued_amount,
                    "repriced_npv": curr_calced_npv,
                    "repriced_ytm": curr_calced_ytm,
                    "price_spread": row[price_col] + curr_accrued_amount - curr_calced_npv,
                    "ytm_spread": (row[yield_col] - curr_calced_ytm) * 100,
                }
            )
        except Exception as e:
            print(row["cusip"], e)

    return pd.DataFrame(bonds)


def reprice_bonds(
    as_of_date: datetime,
    ql_zero_curves: Dict[str, ql.ZeroCurve],
    curve_set_df: pd.DataFrame,
):
    price_cols = ["bid_price", "offer_price", "mid_price", "eod_price"]
    required_cols = ["issue_date", "maturity_date", "int_rate"]
    price_col_exists = any(item in curve_set_df.columns for item in price_cols)
    missing_required_cols = [item for item in required_cols if item not in curve_set_df.columns]

    if not price_col_exists:
        raise ValueError(f"Build Spot Curve - Couldn't find a valid price col in your curve set df - one of {price_cols}")
    if missing_required_cols:
        raise ValueError(f"Build Spot Curve - Missing required curve set cols: {missing_required_cols}")

    price_col = next((item for item in price_cols if item in curve_set_df.columns), None)
    quote_type = price_col.split("_")[0]
    yield_col = f"{quote_type}_yield"

    if yield_col not in curve_set_df.columns:
        raise ValueError(f"Build Spot Curve - Missing required curve set cols: {yield_col}")

    calendar = ql.UnitedStates(m=ql.UnitedStates.GovernmentBond)
    today = calendar.adjust(pydatetime_to_quantlib_date(py_datetime=as_of_date))
    ql.Settings.instance().evaluationDate = today

    t_plus = 1
    bond_settlement_date = calendar.advance(today, ql.Period(t_plus, ql.Days))
    frequency = ql.Semiannual
    day_count = ql.ActualActual(ql.ActualActual.ISDA)
    par = 100.0

    bonds: List[Dict[str, ql.FixedRateBond | ql.ZeroCouponBond]] = []
    for _, row in curve_set_df.iterrows():
        try:
            maturity = pydatetime_to_quantlib_date(row["maturity_date"])
            if np.isnan(row["int_rate"]):
                bond_ql = ql.ZeroCouponBond(
                    t_plus,
                    calendar,
                    par,
                    maturity,
                    ql.ModifiedFollowing,
                    100.0,
                    bond_settlement_date,
                )
                bond_rl: rl.FixedRateBond = rl.Bill(
                    termination=row["maturity_date"],
                    effective=row["issue_date"],
                    calendar="nyc",
                    modifier="NONE",
                    currency="usd",
                    convention="Act360",
                    settle=1,
                    curves="bill_curve",
                    calc_mode="us_gbb",
                )
            else:
                schedule = ql.Schedule(
                    bond_settlement_date,
                    maturity,
                    ql.Period(frequency),
                    calendar,
                    ql.ModifiedFollowing,
                    ql.ModifiedFollowing,
                    ql.DateGeneration.Backward,
                    False,
                )
                bond_ql = ql.FixedRateBond(
                    t_plus,
                    100.0,
                    schedule,
                    [row["int_rate"] / 100],
                    day_count,
                    ql.ModifiedFollowing,
                )
                bond_rl: rl.FixedRateBond = rl.FixedRateBond(
                    effective=row["issue_date"],
                    termination=row["maturity_date"],
                    fixed_rate=row["int_rate"],
                    spec="ust",
                    calc_mode="ust_31bii",
                )

            curr_accrued_amount = bond_rl.accrued(quantlib_date_to_pydatetime(bond_settlement_date))
            curr_row = {
                "cusip": row["cusip"],
                "label": row["label"],
                "issue_date": row["issue_date"],
                "maturity_date": row["maturity_date"],
                "time_to_maturity": row["time_to_maturity"],
                "high_investment_rate": row["high_investment_rate"],
                "int_rate": row["int_rate"],
                "rank": row["rank"] if "rank" in curve_set_df.columns else None,
                "outstanding": (row["outstanding_amt"] if "outstanding_amt" in curve_set_df.columns else None),
                "soma_holdings": (row["parValue"] if "parValue" in curve_set_df.columns else None),
                "stripping_amount": (row["portion_stripped_amt"] if "portion_stripped_amt" in curve_set_df.columns else None),
                "free_float": (row["free_float"] if "free_float" in curve_set_df.columns else None),
                yield_col: row[yield_col],
                price_col: row[price_col],
                "accured": curr_accrued_amount,
            }

            for label, ql_zero_curve in ql_zero_curves.items():
                yield_curve_handle = ql.YieldTermStructureHandle(ql_zero_curve)
                engine = ql.DiscountingBondEngine(yield_curve_handle)
                bond_ql.setPricingEngine(engine)
                curr_calced_npv = bond_ql.NPV()
                curr_calced_ytm = (
                    bond_ql.bondYield(
                        curr_calced_npv,
                        day_count,
                        ql.Compounded,
                        frequency,
                        bond_settlement_date,
                    )
                    * 100
                )

                curr_price_spread = row[price_col] + curr_accrued_amount - curr_calced_npv
                curr_ytm_spread = (row[yield_col] - curr_calced_ytm) * 100

                curr_row[f"{label}_repriced_npv"] = curr_calced_npv
                curr_row[f"{label}_repriced_ytm"] = curr_calced_ytm
                curr_row[f"{label}_price_spread"] = curr_price_spread
                curr_row[f"{label}_ytm_spread"] = curr_ytm_spread

            bonds.append(curr_row)

        except Exception as e:
            print(row["cusip"], e)

    return pd.DataFrame(bonds)


# def reprice_bonds_single_scipy_interpolated_curve(
#     as_of_date: datetime,
#     scipy_interp_curve: scipy.interpolate.interp1d,
#     curve_set_df: pd.DataFrame,
#     yield_col: str,
#     price_col: str,
#     coupon_col: str,
#     mat_col: str,
# ) -> pd.DataFrame:
#     # price_cols = ["bid_price", "offer_price", "mid_price", "eod_price"]
#     # required_cols = ["issue_date", "maturity_date", "int_rate"]
#     # price_col_exists = any(item in curve_set_df.columns for item in price_cols)
#     # missing_required_cols = [
#     #     item for item in required_cols if item not in curve_set_df.columns
#     # ]

#     # if not price_col_exists:
#     #     raise ValueError(
#     #         f"Build Spot Curve - Couldn't find a valid price col in your curve set df - one of {price_cols}"
#     #     )
#     # if missing_required_cols:
#     #     raise ValueError(
#     #         f"Build Spot Curve - Missing required curve set cols: {missing_required_cols}"
#     #     )

#     # price_col = next(
#     #     (item for item in price_cols if item in curve_set_df.columns), None
#     # )
#     # quote_type = price_col.split("_")[0]
#     # yield_col = f"{quote_type}_yield"

#     # if yield_col not in curve_set_df.columns:
#     #     raise ValueError(
#     #         f"Build Spot Curve - Missing required curve set cols: {yield_col}"
#     #     )

#     as_of_date = np.datetime64(as_of_date)
#     t_plus = 1
#     bond_settlement_date = as_of_date + BDay(t_plus)
#     frequency = 2
#     par = 100.0

#     bonds: List[Dict[str, float]] = []
#     for _, row in curve_set_df.iterrows():
#         try:
#             maturity_date = np.datetime64(row[mat_col])
#             time_to_maturity = (maturity_date - bond_settlement_date).days / 365
#             discount_factor = np.exp(
#                 -scipy_interp_curve(time_to_maturity) * time_to_maturity
#             )

#             if np.isnan(row[coupon_col]):
#                 curr_calced_npv = par * discount_factor
#             else:
#                 num_periods = int(frequency * time_to_maturity)
#                 cash_flows = [
#                     (par * (row[coupon_col] / 100) / frequency) * discount_factor**i
#                     for i in range(1, num_periods + 1)
#                 ]
#                 cash_flows.append(par * discount_factor)
#                 curr_calced_npv = sum(cash_flows)

#             curr_calced_ytm = -np.log(curr_calced_npv / par) / time_to_maturity * 100
#             bonds.append(
#                 {
#                     "label": row["label"],
#                     yield_col: row[yield_col],
#                     price_col: row[price_col],
#                     "repriced_npv": curr_calced_npv,
#                     "repriced_ytm": curr_calced_ytm,
#                     "price_spread": row[price_col] - curr_calced_npv,
#                     "ytm_spread": (row[yield_col] - curr_calced_ytm) * 100,
#                 }
#             )
#         except Exception as e:
#             print(row["label"], e)

#     return pd.DataFrame(bonds)


def calc_ust_metrics(
    bond_info: str, 
    curr_price: float, 
    curr_ytm: float, 
    on_rate: float,
    as_of_date: datetime, 
    scipy_interp: scipy.interpolate.interpolate, 
    print_bond_info=False, 
):
    calendar = ql.UnitedStates(ql.UnitedStates.GovernmentBond)
    day_count = ql.ActualActual(ql.ActualActual.ISDA)
    times = np.arange(0, 30.5, 0.5)

    dates = []
    zero_rates = []
    today = calendar.adjust(pydatetime_to_quantlib_date(py_datetime=as_of_date))
    ql.Settings.instance().evaluationDate = today
    t_plus = 2
    bond_settlement_date = calendar.advance(today, ql.Period(t_plus, ql.Days))

    for t in times:
        if t == 0:
            dates.append(today)
            zero_rate = scipy_interp(0)
            zero_rates.append(float(zero_rate) / 100)
        else:
            maturity_date = calendar.advance(pydatetime_to_quantlib_date(as_of_date), ql.Period(int(round(t * 365)), ql.Days))
            dates.append(maturity_date)
            zero_rate = scipy_interp(t)
            zero_rates.append(float(zero_rate) / 100)

    zero_curve = ql.ZeroCurve(dates, zero_rates, day_count, calendar)
    yield_curve_handle = ql.YieldTermStructureHandle(zero_curve)
    engine = ql.DiscountingBondEngine(yield_curve_handle)

    ql_fixed_rate_bond_obj = to_quantlib_fixed_rate_bond_obj(bond_info=bond_info, as_of_date=as_of_date, print_bond_info=print_bond_info)
    ql_fixed_rate_bond_obj.setPricingEngine(engine)

    try:
        zspread = ql.BondFunctions.zSpread(
            ql_fixed_rate_bond_obj,
            curr_price,
            yield_curve_handle.currentLink(),
            day_count,
            ql.Compounded,
            ql.Semiannual,
            pydatetime_to_quantlib_date(as_of_date),
            1.0e-16,
            1000000,
            0.0,
        )
        spread1 = ql.SimpleQuote(zspread)
        spread_handle1 = ql.QuoteHandle(spread1)
        ts_spreaded1 = ql.ZeroSpreadedTermStructure(yield_curve_handle, spread_handle1, ql.Compounded, ql.Semiannual)
        ts_spreaded_handle1 = ql.YieldTermStructureHandle(ts_spreaded1)
        ycsin = ts_spreaded_handle1
        bond_engine = ql.DiscountingBondEngine(ycsin)
        ql_fixed_rate_bond_obj.setPricingEngine(bond_engine)
        zspread_impl_clean_price = ql_fixed_rate_bond_obj.cleanPrice()
        zspread = zspread * 10000
    except:
        zspread = None
        zspread_impl_clean_price = None

    rate = ql.InterestRate(curr_ytm / 100, day_count, ql.Compounded, ql.Semiannual)
    bps_value = ql.BondFunctions.basisPointValue(ql_fixed_rate_bond_obj, rate, bond_settlement_date)
    dv01_1mm = bps_value * 1_000_000 / 100
    impl_spot_3m_fwds = calc_ust_impl_spot_n_fwd_curve(n=0.25, scipy_interp_curve=scipy_interp, return_scipy=True)
    impl_spot_6m_fwds = calc_ust_impl_spot_n_fwd_curve(n=0.5, scipy_interp_curve=scipy_interp, return_scipy=True)
    impl_spot_12m_fwds = calc_ust_impl_spot_n_fwd_curve(n=1, scipy_interp_curve=scipy_interp, return_scipy=True)
    bond_ttm: timedelta = (bond_info["maturity_date"] - as_of_date)

    metrics = {
        "Date": as_of_date,
        "zspread": zspread,
        "zspread_impl_clean_price": zspread_impl_clean_price,
        "clean_price": ql.BondFunctions.cleanPrice(ql_fixed_rate_bond_obj, rate),
        "dirty_price": ql.BondFunctions.dirtyPrice(ql_fixed_rate_bond_obj, yield_curve_handle.currentLink(), bond_settlement_date),
        "accrued_amount": ql.BondFunctions.accruedAmount(ql_fixed_rate_bond_obj, bond_settlement_date),
        "bps": bps_value,
        "dv01_1mm": dv01_1mm,
        "mac_duration": ql.BondFunctions.duration(ql_fixed_rate_bond_obj, rate, ql.Duration.Macaulay),
        "mod_duration": ql.BondFunctions.duration(ql_fixed_rate_bond_obj, rate, ql.Duration.Modified),
        "convexity": ql.BondFunctions.convexity(ql_fixed_rate_bond_obj, rate, bond_settlement_date),
        "basis_point_value": ql.BondFunctions.basisPointValue(ql_fixed_rate_bond_obj, rate, bond_settlement_date),
        "yield_value_basis_point": ql.BondFunctions.yieldValueBasisPoint(ql_fixed_rate_bond_obj, rate, bond_settlement_date),
        "rough_carry": curr_ytm - on_rate, 
        "rough_3m_rolldown": (impl_spot_3m_fwds(float(bond_ttm.days / 365)) - curr_ytm) * 100,
        "rough_6m_rolldown": (impl_spot_6m_fwds(float(bond_ttm.days / 365)) - curr_ytm) * 100,
        "rough_12m_rolldown": (impl_spot_12m_fwds(float(bond_ttm.days / 365)) - curr_ytm) * 100
    }

    return metrics


def calc_ust_impl_spot_n_fwd_curve(n: float | int, scipy_interp_curve: scipy.interpolate.interpolate, return_scipy=False) -> Dict[float, float]:
    cfs = np.arange(0.5, 30 + 1, 0.5)
    implied_spot_rates = []
    first_n_cfs = 0
    for t in cfs:
        if t > n:
            Z_t_temp = scipy_interp_curve(t)
            Z_n = scipy_interp_curve(n)
            Z_n_t = (Z_t_temp * t - Z_n * n) / (t - n)
            implied_spot_rates.append(Z_n_t)
        else:
            if return_scipy:
                # implied_spot_rates.append(0)
                first_n_cfs += 1
            else:
                implied_spot_rates.append(np.nan)
    
    if return_scipy:
        return scipy.interpolate.CubicSpline(x=cfs[first_n_cfs:], y=implied_spot_rates)
    return dict(zip(cfs, implied_spot_rates))        


def calc_ust_metrics_parallel(
    bonds: List[Dict],
    curve_data_fetcher: CurveDataFetcher,
    inter_func_key: str,
    max_workers=16,
) -> List[Dict]:

    results = []

    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        future_to_args = {
            executor.submit(
                calc_ust_metrics,
                curve_data_fetcher.ust_data_fetcher.cusip_to_ust_label(cusip=bond["cusip"]),
                bond["eod_price"],
                bond["eod_yield"],
                bond["Date"],
                bond[inter_func_key],
            ): (curve_data_fetcher.ust_data_fetcher.cusip_to_ust_label(cusip=bond["cusip"]), bond["Date"])
            for bond in bonds
        }

        for future in as_completed(future_to_args):
            bond_info, as_of_date = future_to_args[future]
            try:
                result = future.result()
                results.append(result)
            except Exception as exc:
                print(f"Bond {bond_info} for date {as_of_date} generated an exception: {exc}")

    return results



# from datetime import datetime
# import pandas as pd
# import ujson as json
# import asyncio
# import warnings
# from datetime import datetime
# from typing import Dict, List, Optional, Tuple
# from requests.models import PreparedRequest

# import httpx
# import pandas as pd

# from DataFetcher.base import DataFetcherBase
# warnings.simplefilter(action="ignore", category=FutureWarning)

# import sys

# if sys.platform == "win32":
#     loop = asyncio.ProactorEventLoop()
#     asyncio.set_event_loop(loop)


# class BarChartDataFetcher(DataFetcherBase):
#     def __init__(
#         self,
#         global_timeout: int = 10,
#         proxies: Optional[Dict[str, str]] = None,
#         debug_verbose: Optional[bool] = False,
#         info_verbose: Optional[bool] = False,
#         error_verbose: Optional[bool] = False,
#     ):
#         super().__init__(
#             global_timeout=global_timeout,
#             proxies=proxies,
#             debug_verbose=debug_verbose,
#             info_verbose=info_verbose,
#             error_verbose=error_verbose,
#         )

#     async def _fetch_timeseries(
#         self,
#         client: httpx.AsyncClient,
#         wsj_ticker_key: str,
#         start_date: Optional[datetime] = None,
#         end_date: Optional[datetime] = None,
#         max_retries: Optional[int] = 3,
#         backoff_factor: Optional[int] = 1,
#         uid: Optional[str | int] = None,
#     ):
#         payload = {
#             "Step": "P1D",
#             "TimeFrame": "all",
#             "EntitlementToken": "57494d5ed7ad44af85bc59a51dd87c90",
#             "IncludeMockTick": True,
#             "FilterNullSlots": True,
#             "FilterClosedPoints": True,
#             "IncludeClosedSlots": True,
#             "IncludeOfficialClose": True,
#             "InjectOpen": True,
#             "ShowPreMarket": True,
#             "ShowAfterHours": True,
#             "UseExtendedTimeFrame": True,
#             "WantPriorClose": True,
#             "IncludeCurrentQuotes": True,
#             "ResetTodaysAfterHoursPercentChange": False,
#             "Series": [
#                 {
#                     "Key": wsj_ticker_key,
#                     "Dialect": "Charting",
#                     "Kind": "Ticker",
#                     "SeriesId": "s1",
#                     "DataTypes": ["Last"],
#                 }
#             ],
#         }
#         params = {
#             "json": json.dumps(payload),
#             "ckey": "57494d5ed7",
#         }
#         url = "https://api.wsj.net/api/michelangelo/timeseries/history"
#         prep_url = PreparedRequest()
#         prep_url.prepare_url(url, params)
#         headers = {
#             "authority": "api.wsj.net",
#             "method": "GET",
#             "path": prep_url.path_url,
#             "scheme": "https",
#             "Connection": "keep-alive",
#             "Pragma": "no-cache",
#             "Cache-Control": "no-cache",
#             "sec-ch-ua": '" Not A;Brand";v="99", "Chromium";v="99", "Google Chrome";v="99"',
#             "Accept": "application/json, text/javascript, */*; q=0.01",
#             "Dylan2010.EntitlementToken": "57494d5ed7ad44af85bc59a51dd87c90",
#             "sec-ch-ua-mobile": "?0",
#             "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.74 Safari/537.36",
#             "sec-ch-ua-platform": '"macOS"',
#             "Origin": "https://www.wsj.com",
#             "Sec-Fetch-Site": "cross-site",
#             "Sec-Fetch-Mode": "cors",
#             "Sec-Fetch-Dest": "empty",
#             "Referer": "https://www.wsj.com/",
#             "Accept-Language": "en-US,en;q=0.9,es;q=0.8",
#         }

#         cols_to_return = ["Date", wsj_ticker_key]
#         retries = 0
#         try:
#             while retries < max_retries:
#                 try:
#                     response = await client.get(prep_url.url, headers=headers)
#                     response.raise_for_status()
#                     json_data = response.json()
#                     df = pd.DataFrame({
#                         "Date": json_data["TimeInfo"]["Ticks"],
#                         wsj_ticker_key:[d[0] for d in json_data["Series"][0]["DataPoints"]]
#                     })
#                     df["Date"] = pd.to_datetime(df["Date"], unit="ms")
#                     if start_date:
#                         df = df[df["Date"] >= start_date]
#                     if end_date:
#                         df = df[df["Date"] <= end_date]
                        
#                     if uid:
#                         return wsj_ticker_key, df, uid
#                     return wsj_ticker_key, df
                    
#                 except httpx.HTTPStatusError as e:
#                     self._logger.error(
#                         f"WSJ - Bad Status: {response.status_code}"
#                     )
#                     if response.status_code == 404:
#                         if uid:
#                             return wsj_ticker_key, pd.DataFrame(columns=cols_to_return), uid
#                         return wsj_ticker_key, pd.DataFrame(columns=cols_to_return)

#                     retries += 1
#                     wait_time = backoff_factor * (2 ** (retries - 1))
#                     self._logger.debug(
#                         f"WSJ- Throttled for {wsj_ticker_key}. Waiting for {wait_time} seconds before retrying..."
#                     )
#                     await asyncio.sleep(wait_time)

#                 except Exception as e:
#                     self._logger.error(f"WSJ - Error: {str(e)}")
#                     retries += 1
#                     wait_time = backoff_factor * (2 ** (retries - 1))
#                     self._logger.debug(
#                         f"WSJ - Throttled for {wsj_ticker_key}. Waiting for {wait_time} seconds before retrying..."
#                     )
#                     await asyncio.sleep(wait_time)

#             raise ValueError(f"WSJ  - Max retries exceeded for {wsj_ticker_key}")

#         except Exception as e:
#             self._logger.error(e)
#             if uid:
#                 return wsj_ticker_key, pd.DataFrame(columns=cols_to_return), uid
#             return wsj_ticker_key, pd.DataFrame(columns=cols_to_return)

#     async def _fetch_timeseries_with_semaphore(
#         self, semaphore, *args, **kwargs
#     ):
#         async with semaphore:
#             return await self._fetch_timeseries(*args, **kwargs)

#     def wsj_timeseries_api(
#         self,
#         wsj_ticker_keys: List[str],
#         start_date: Optional[datetime] = None,
#         end_date: Optional[datetime] = None,
#         max_concurrent_tasks: int = 64,
#     ):
#         async def build_tasks(
#             client: httpx.AsyncClient,
#             wsj_ticker_keys: List[str],
#             start_date: datetime,
#             end_date: datetime,
#         ):
#             semaphore = asyncio.Semaphore(max_concurrent_tasks)
#             tasks = [
#                 self._fetch_timeseries_with_semaphore(
#                     semaphore=semaphore,
#                     client=client,
#                     wsj_ticker_key=wsj_ticker_key,
#                     start_date=start_date,
#                     end_date=end_date,
#                 )
#                 for wsj_ticker_key in wsj_ticker_keys
#             ]
#             return await asyncio.gather(*tasks)

#         async def run_fetch_all(
#             wsj_ticker_keys: List[str],
#             start_date: datetime,
#             end_date: datetime,
#         ):
#             async with httpx.AsyncClient(proxy=self._proxies["https"]) as client:
#                 all_data = await build_tasks(
#                     client=client,
#                     wsj_ticker_keys=wsj_ticker_keys,
#                     start_date=start_date,
#                     end_date=end_date,
#                 )
#                 return all_data

#         dfs: List[Tuple[str, pd.DataFrame]] = asyncio.run(
#             run_fetch_all(
#                 wsj_ticker_keys=wsj_ticker_keys,
#                 start_date=start_date,
#                 end_date=end_date,
#             )
#         )
#         return dict(dfs)


#!/usr/bin/env python3

# SPIA - Income annuity (SPIA and DIA) price calculator
# Copyright (C) 2018 Gordon Irlam
#
# This program may be licensed by you (at your option) under an Open
# Source, Free for Non-Commercial Use, or Commercial Use License.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is free for non-commercial use: you can use and modify it
# under the terms of the Creative Commons
# Attribution-NonCommercial-ShareAlike 4.0 International Public License
# (https://creativecommons.org/licenses/by-nc-sa/4.0/).
#
# A Commercial Use License is available in exchange for agreed
# remuneration.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class MonotoneConvex(object):
    """Hagan, Patrick S., and Graeme West. "Methods for constructing a yield curve." Wilmott Magazine, May (2008): 70-81."""

    def collar(self, a, b, c):

        return max(a, min(b, c))

    def last_index(self, term):

        for i in range(len(self.terms) - 1):
            if term < self.terms[i + 1]:
                return i

        return len(self.terms) - 2

    def __init__(
        self,
        terms,
        spots,
        *,
        min_long_term_forward=float("inf"),
        force_forwards_non_negative=True
    ):
        """Initialize a monotone convex object.

        term is an ascending list of bond terms in years, exluding
        zero.

        spots is the corresponding list of continuously compounding
        annualized spot rates.

        min_long_term_forward is the minimum term to use in computing
        the average forward rate. The average forward rate is used as
        the forward rate for terms beyond on the longest term
        specified. If min_long_term_forward exceeds the longest term
        the forward rate of the longest term is used for the forward
        rate of terms beyond the longest term. Set
        min_long_term_forward to 15 to use the Treasury methodology,
        which is to use for the extrapolated forward rate the average
        observed forward rate beyond 15 years. Setting
        min_long_term_forward to a value less than the longest term
        may introduce a discontinuity in forward rates at the longest
        term which may not be desired.

        force_forwards_non_negative specifies whether forward rates
        are forced to be non-negative. If so the negative derivative
        of the spot rate determined price need not match the forward
        rate.

        """

        self.force_forwards_non_negative = force_forwards_non_negative

        assert len(terms) == len(spots)

        # extend the curve to time 0
        terms = [0] + list(terms)
        spots = [spots[0]] + list(spots)

        # step 1
        fdiscrete = []
        for j in range(len(terms) - 1):
            fdiscrete.append(
                (terms[j + 1] * spots[j + 1] - terms[j] * spots[j])
                / (terms[j + 1] - terms[j])
            )

        # step 2
        f = [None]
        for j in range(len(terms) - 2):
            f.append(
                (terms[j + 1] - terms[j])
                / (terms[j + 1 + 1] - terms[j])
                * fdiscrete[j + 1]
                + (terms[j + 2] - terms[j + 1])
                / (terms[j + 2] - terms[j])
                * fdiscrete[j]
            )

        # step 3
        f[0] = fdiscrete[0] - 0.5 * (f[1] - fdiscrete[0])
        f.append(fdiscrete[-1] - 0.5 * (f[-1] - fdiscrete[-1]))
        if force_forwards_non_negative:
            f[0] = self.collar(0, f[0], 2 * fdiscrete[0])
            f[-1] = self.collar(0, f[-1], 2 * fdiscrete[-1])
            for j in range(len(terms) - 2):
                f[j + 1] = self.collar(
                    0, f[j + 1], 2 * min(fdiscrete[j], fdiscrete[j + 1])
                )

        self.terms = tuple(terms)
        self.spots = tuple(spots)
        self.fdiscrete = tuple(fdiscrete)
        self.f = tuple(f)

        if min_long_term_forward < terms[-1]:
            self.long_term_forward = (
                terms[-1] * self.spot(terms[-1])
                - min_long_term_forward * self.spot(min_long_term_forward)
            ) / (terms[-1] - min_long_term_forward)
        else:
            self.long_term_forward = self.f[-1]

    def forward(self, term):
        """Return the continuously compounding annualized forward rate for
        term, term.

        """

        if term <= 0:
            forward = self.f[0]
        elif term == self.terms[-1]:
            forward = self.f[-1]
        elif term > self.terms[-1]:
            forward = self.long_term_forward
        else:
            i = self.last_index(term)
            # the x in (25)
            x = (term - self.terms[i]) / (self.terms[i + 1] - self.terms[i])
            g0 = self.f[i] - self.fdiscrete[i]
            g1 = self.f[i + 1] - self.fdiscrete[i]
            if x == 0:
                g = g0
            elif x == 1:
                g = g1
            elif (g0 < 0 and -0.5 * g0 <= g1 and g1 <= -2 * g0) or (
                g0 > 0 and -0.5 * g0 >= g1 and g1 >= -2 * g0
            ):
                # zone (i)
                g = g0 * (1 - 4 * x + 3 * x**2) + g1 * (-2 * x + 3 * x**2)
            elif (g0 < 0 and g1 > -2 * g0) or (g0 > 0 and g1 < -2 * g0):
                # zone (ii)
                # (29)
                eta = (g1 + 2 * g0) / (g1 - g0)
                # (28)
                if x <= eta:
                    g = g0
                else:
                    g = g0 + (g1 - g0) * ((x - eta) / (1 - eta)) ** 2
            elif (g0 > 0 and 0 > g1 and g1 > -0.5 * g0) or (
                g0 < 0 and 0 < g1 and g1 < -0.5 * g0
            ):
                # zone (iii)
                # (31)
                eta = 3 * g1 / (g1 - g0)
                # (30)
                if x < eta:
                    g = g1 + (g0 - g1) * ((eta - x) / eta) ** 2
                else:
                    g = g1
            elif g0 == 0 and g1 == 0:
                g = 0
            else:
                # zone (iv)
                # (33)
                eta = g1 / (g1 + g0)
                # (34)
                a = -g0 * g1 / (g0 + g1)
                # (32)
                if x <= eta:
                    g = a + (g0 - a) * ((eta - x) / eta) ** 2
                else:
                    g = a + (g1 - a) * ((eta - x) / (1 - eta)) ** 2
                    # (26)
            forward = g + self.fdiscrete[i]

        if self.force_forwards_non_negative:
            forward = max(0, forward)

        return forward

    def spot(self, term):
        """Return the continuously compounding annualized spot rate for term,
        term.

        """

        if term <= 0:
            return self.f[0]
        elif term > self.terms[-1]:
            return (
                self.terms[-1] * self.spots[-1]
                + (term - self.terms[-1]) * self.long_term_forward
            ) / term
        else:
            i = self.last_index(term)
            l = self.terms[i + 1] - self.terms[i]
            # the x in (25)
            x = (term - self.terms[i]) / l
            g0 = self.f[i] - self.fdiscrete[i]
            g1 = self.f[i + 1] - self.fdiscrete[i]
            if x == 0 or x == 1:
                g = 0
            elif (g0 < 0 and -0.5 * g0 <= g1 and g1 <= -2 * g0) or (
                g0 > 0 and -0.5 * g0 >= g1 and g1 >= -2 * g0
            ):
                # zone [i]
                g = l * (g0 * (x - 2 * x**2 + x**3) + g1 * (-(x**2) + x**3))
            elif (g0 < 0 and g1 > -2 * g0) or (g0 > 0 and g1 < -2 * g0):
                # zone (ii)
                # (29)
                eta = (g1 + 2 * g0) / (g1 - g0)
                # (28)
                if x <= eta:
                    g = g0 * (term - self.terms[i])
                else:
                    g = (
                        g0 * (term - self.terms[i])
                        + (g1 - g0) * (x - eta) ** 3 / (1 - eta) ** 2 / 3 * l
                    )
            elif (g0 > 0 and 0 > g1 and g1 > -0.5 * g0) or (
                g0 < 0 and 0 < g1 and g1 < -0.5 * g0
            ):
                # zone (iii)
                # (31)
                eta = 3 * g1 / (g1 - g0)
                # (30)
                if x < eta:
                    g = l * (
                        g1 * x - 1 / 3 * (g0 - g1) * ((eta - x) ** 3 / eta**2 - eta)
                    )
                else:
                    g = l * (2 / 3 * g1 + 1 / 3 * g0) * eta + g1 * (x - eta) * l
            elif g0 == 0 and g1 == 0:
                g = 0
            else:
                # zone (iv)
                # (33)
                eta = g1 / (g1 + g0)
                # (34)
                a = -g0 * g1 / (g0 + g1)
                # (32)
                if x <= eta:
                    g = l * (a * x - 1 / 3 * (g0 - a) * ((eta - x) ** 3 / eta**2 - eta))
                else:
                    g = l * (2 / 3 * a + 1 / 3 * g0) * eta + l * (
                        a * (x - eta) + (g1 - a) / 3 * (x - eta) ** 3 / (1 - eta) ** 2
                    )
                    # (12)
            return (
                self.terms[i] * self.spots[i]
                + self.fdiscrete[i] * (term - self.terms[i])
                + g
            ) / term



# adapted from https://github.com/luphord/nelson_siegel_svensson/blob/master/nelson_siegel_svensson/calibrate.py
# -*- coding: utf-8 -*-

from typing import Any, Optional, Tuple

import numpy as np
import numpy.typing as npt
import pandas as pd
from numpy.linalg import lstsq
from scipy.optimize import OptimizeResult, minimize

from CurvyCUSIPs.models.BjorkChristensen import BjorkChristensenCurve
from CurvyCUSIPs.models.BjorkChristensenAugmented import BjorkChristensenAugmentedCurve
from CurvyCUSIPs.models.DieboldLi import DieboldLiCurve
from CurvyCUSIPs.models.MLESM import MerrillLynchExponentialSplineModel
from CurvyCUSIPs.models.NelsonSiegel import NelsonSiegelCurve
from CurvyCUSIPs.models.NelsonSiegelSvensson import NelsonSiegelSvenssonCurve
from CurvyCUSIPs.models.SmithWilson import SmithWilsonCurve, find_ufr_ytm


def _assert_same_shape(t: np.ndarray, y: np.ndarray) -> None:
    assert t.shape == y.shape, "Mismatching shapes of time and values"


def betas_ns_ols(tau: float, t: np.ndarray, y: np.ndarray) -> Tuple[NelsonSiegelCurve, Any]:
    _assert_same_shape(t, y)
    curve = NelsonSiegelCurve(0, 0, 0, tau)
    factors = curve.factor_matrix(t)
    lstsq_res = lstsq(factors, y, rcond=None)
    beta = lstsq_res[0]
    return NelsonSiegelCurve(beta[0], beta[1], beta[2], tau), lstsq_res


def errorfn_ns_ols(tau: float, t: np.ndarray, y: np.ndarray) -> float:
    _assert_same_shape(t, y)
    curve, lstsq_res = betas_ns_ols(tau, t, y)
    return np.sum((curve(t) - y) ** 2)


def calibrate_ns_ols(t: np.ndarray, y: np.ndarray, tau0: float = 2.0) -> Tuple[NelsonSiegelCurve, Any]:
    _assert_same_shape(t, y)
    opt_res = minimize(errorfn_ns_ols, x0=tau0, args=(t, y))
    curve, lstsq_res = betas_ns_ols(opt_res.x[0], t, y)
    return curve, opt_res


def empirical_factors(y_3m: float, y_2y: float, y_10y: float) -> Tuple[float, float, float]:
    return y_10y, y_10y - y_3m, 2 * y_2y - y_3m - y_10y


def betas_nss_ols(tau: Tuple[float, float], t: np.ndarray, y: np.ndarray) -> Tuple[NelsonSiegelSvenssonCurve, Any]:
    _assert_same_shape(t, y)
    curve = NelsonSiegelSvenssonCurve(0, 0, 0, 0, tau[0], tau[1])
    factors = curve.factor_matrix(t)
    lstsq_res = lstsq(factors, y, rcond=None)
    beta = lstsq_res[0]
    return (
        NelsonSiegelSvenssonCurve(beta[0], beta[1], beta[2], beta[3], tau[0], tau[1]),
        lstsq_res,
    )


def errorfn_nss_ols(tau: Tuple[float, float], t: np.ndarray, y: np.ndarray) -> float:
    _assert_same_shape(t, y)
    curve, lstsq_res = betas_nss_ols(tau, t, y)
    return np.sum((curve(t) - y) ** 2)


def calibrate_nss_ols(t: np.ndarray, y: np.ndarray, tau0: Tuple[float, float] = (2.0, 5.0)) -> Tuple[NelsonSiegelSvenssonCurve, Any]:
    _assert_same_shape(t, y)
    opt_res = minimize(errorfn_nss_ols, x0=np.array(tau0), args=(t, y))
    curve, lstsq_res = betas_nss_ols(opt_res.x, t, y)
    return curve, opt_res, lstsq_res


def betas_nss_weighted_ols(tau: Tuple[float, float], t: np.ndarray, y: np.ndarray, weights: np.ndarray) -> Tuple[NelsonSiegelSvenssonCurve, Any]:
    """
    Modified OLS fitting to include weights for the weighted Svensson model.
    """
    _assert_same_shape(t, y, weights)  # Ensure that all inputs have the same shape
    curve = NelsonSiegelSvenssonCurve(0, 0, 0, 0, tau[0], tau[1])
    factors = curve.factor_matrix(t)

    # Apply the weights to the factors and yields
    W = np.diag(weights)
    weighted_factors = W @ factors
    weighted_y = W @ y

    # Solve the weighted OLS problem
    lstsq_res = lstsq(weighted_factors, weighted_y, rcond=None)
    beta = lstsq_res[0]

    return (
        NelsonSiegelSvenssonCurve(beta[0], beta[1], beta[2], beta[3], tau[0], tau[1]),
        lstsq_res,
    )


def weighted_errorfn_nss_ols(tau: Tuple[float, float], t: np.ndarray, y: np.ndarray, weights: np.ndarray) -> float:
    """
    Error function for the weighted Svensson model to be used in optimization.
    """
    curve, lstsq_res = betas_nss_weighted_ols(tau, t, y, weights)
    return np.sum(weights * (curve(t) - y) ** 2)


def calibrate_nss_weighted_ols(
    t: np.ndarray, y: np.ndarray, weights: np.ndarray, tau0: Tuple[float, float] = (2.0, 5.0)
) -> Tuple[NelsonSiegelSvenssonCurve, Any]:
    """
    Calibration function for the weighted Svensson model using OLS.
    """
    _assert_same_shape(t, y, weights)
    opt_res = minimize(weighted_errorfn_nss_ols, x0=np.array(tau0), args=(t, y, weights))
    curve, lstsq_res = betas_nss_weighted_ols(opt_res.x, t, y, weights)
    return curve, opt_res, lstsq_res


def errorfn_bc_ols(tau: float, t: np.ndarray, y: np.ndarray) -> float:
    curve, _ = betas_bc_ols(tau, t, y)
    estimated_yields = curve(t)
    return np.sum((estimated_yields - y) ** 2)


def betas_bc_ols(tau: float, t: np.ndarray, y: np.ndarray) -> Tuple[BjorkChristensenCurve, np.linalg.LinAlgError]:
    curve = BjorkChristensenCurve(0, 0, 0, 0, tau)
    F = curve.factor_matrix(t)
    betas, _, _, _ = np.linalg.lstsq(F, y, rcond=None)
    curve.beta0, curve.beta1, curve.beta2, curve.beta3 = betas
    return curve, betas


def calibrate_bc_ols(t: np.ndarray, y: np.ndarray, tau0: float = 1.0) -> Tuple[BjorkChristensenCurve, Any]:
    _assert_same_shape(t, y)
    opt_res = minimize(errorfn_bc_ols, x0=np.array([tau0]), args=(t, y))
    curve, lstsq_res = betas_bc_ols(opt_res.x[0], t, y)
    return curve, opt_res


def calibrate_bc_augmented_ols(maturities: npt.NDArray[np.float64], yields: npt.NDArray[np.float64]) -> Tuple[BjorkChristensenAugmentedCurve, Any]:

    def objective(params: npt.NDArray[np.float64]) -> float:
        beta0, beta1, beta2, beta3, beta4, tau = params
        curve = BjorkChristensenAugmentedCurve(beta0, beta1, beta2, beta3, beta4, tau)
        model_yields = curve(np.array(maturities))
        return np.sum((model_yields - np.array(yields)) ** 2)

    initial_params = [0.01, 0.01, 0.01, 0.01, 0.01, 1.0]
    bounds = [
        (0, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (0, None),
    ]

    result = minimize(objective, initial_params, method="L-BFGS-B", bounds=bounds)
    fitted_params = result.x
    curve = BjorkChristensenAugmentedCurve(*fitted_params)
    return curve, result


def calibrate_diebold_li_ols(maturities: npt.NDArray[np.float64], yields: npt.NDArray[np.float64]) -> Tuple[DieboldLiCurve, Any]:
    initial_guess = [np.mean(yields), -0.02, 0.02, 0.1]

    def objective(params: npt.NDArray[np.float64]) -> float:
        curve = DieboldLiCurve(params[0], params[1], params[2], params[3])
        return np.sum((curve(np.array(maturities)) - np.array(yields)) ** 2)

    result = minimize(objective, initial_guess, method="BFGS")

    optimized_params = result.x
    curve = DieboldLiCurve(*optimized_params)
    return curve, result


def calibrate_mles_ols(
    maturities: npt.NDArray[np.float64],
    yields: npt.NDArray[np.float64],
    N: int = 8,
    regularization: float = 1e-4,
    overnight_rate: Optional[float] = None,
) -> Tuple[MerrillLynchExponentialSplineModel, Any]:
    if maturities[0] != 0:
        short_rate = overnight_rate or yields[0]
        maturities = np.insert(maturities, 0, 1 / 365)
        yields = np.insert(yields, 0, short_rate)

    """Fit the MLES model to the given yields using OLS."""
    initial_guess = [0.1] + [1.0] * N

    def objective(params: npt.NDArray[np.float64]) -> float:
        alpha = params[0]
        lambda_hat = np.array(params[1:])
        curve = MerrillLynchExponentialSplineModel(alpha, N, lambda_hat)
        curve.fit(np.array(maturities), np.array(yields), np.eye(len(maturities)))
        theoretical_yields = curve.theoretical_yields(np.array(maturities))
        regularization_term = regularization * np.sum(np.diff(lambda_hat) ** 2)

        return np.sum((theoretical_yields - np.array(yields)) ** 2) + regularization_term

    result = minimize(objective, initial_guess, method="BFGS")
    optimized_params = result.x
    optimized_alpha = optimized_params[0]
    optimized_lambda_hat = np.array(optimized_params[1:])
    curve = MerrillLynchExponentialSplineModel(optimized_alpha, N, optimized_lambda_hat)
    curve.fit(np.array(maturities), np.array(yields), np.eye(len(maturities)))

    return curve, result


def calibrate_smith_wilson_ols(
    maturities: npt.NDArray[np.float64],
    yields: npt.NDArray[np.float64],
    ufr: Optional[float] = None,
    alpha_initial: float = 0.1,
    overnight_rate: Optional[float] = None,
) -> Tuple[MerrillLynchExponentialSplineModel, Any]:
    if maturities[0] != 0:
        short_rate = overnight_rate or yields[0]
        maturities = np.insert(maturities, 0, 1 / 365)
        yields = np.insert(yields, 0, short_rate)

    if not ufr:
        ufr = find_ufr_ytm(maturities=maturities, ytms=yields)

    def objective(alpha: float) -> float:
        curve = SmithWilsonCurve(ufr, alpha)
        curve.fit(yields, maturities)
        fitted_yields = curve(maturities)
        return np.sum((fitted_yields - yields) ** 2)

    result = minimize(objective, x0=alpha_initial, bounds=[(0.01, 1.0)], method="L-BFGS-B")

    optimal_alpha = result.x[0]
    calibrated_curve = SmithWilsonCurve(ufr, optimal_alpha)
    calibrated_curve.fit(yields, maturities)

    return calibrated_curve, result


# def calibrate_pca_yield_curve(
#     ytms: npt.NDArray[np.float64], historical_df: pd.DataFrame, n_components: int = 3, use_changes: bool = False
# ) -> Tuple[PCACurve, Any]:
#     if use_changes:
#         historical_df = historical_df.diff().dropna()
#     pca_model = PCACurve(n_components=n_components)
#     pca_model.fit(ytms)
#     return pca_model, pca_model.explained_variance



import asyncio
import re
import warnings
from datetime import datetime
from typing import Callable, Dict, List, Literal, Optional, Tuple

import httpx
import numpy as np
import pandas as pd
import requests
import scipy
import scipy.interpolate

from CurvyCUSIPs.DataFetcher.base import DataFetcherBase

warnings.filterwarnings("ignore", category=pd.errors.SettingWithCopyWarning)
warnings.simplefilter(action="ignore", category=FutureWarning)

import sys

if sys.platform == "win32":
    loop = asyncio.ProactorEventLoop()
    asyncio.set_event_loop(loop)


class BBGSEF_DataFetcher(DataFetcherBase):
    def __init__(
        self,
        global_timeout: int = 10,
        proxies: Optional[Dict[str, str]] = None,
        debug_verbose: Optional[bool] = False,
        info_verbose: Optional[bool] = False,
        error_verbose: Optional[bool] = False,
    ):
        super().__init__(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )
        self.bbg_sef_url = "https://data.bloombergsef.com/bas/blotdatasvc"
        self.bbg_sef_headers = {
            "Accept": "text/plain, */*; q=0.01",
            "Accept-Encoding": "gzip, deflate, br, zstd",
            "Accept-Language": "en-US,en;q=0.9",
            "Cache-Control": "no-cache",
            "Connection": "keep-alive",
            "Content-Type": "application/json",
            "Cookie": "session=eyJmbGFzaCI6e319; session.sig=H0ndcv_UJa_BjmoKTVIBrrCbVpQ",
            "DNT": "1",
            "Host": "data.bloombergsef.com",
            "Origin": "https://data.bloombergsef.com",
            "Pragma": "no-cache",
            "Referer": "https://data.bloombergsef.com/",
            "Sec-CH-UA": '"Chromium";v="130", "Google Chrome";v="130", "Not?A_Brand";v="99"',
            "Sec-CH-UA-Mobile": "?0",
            "Sec-CH-UA-Platform": '"Windows"',
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-origin",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/130.0.0.0 Safari/537.36",
            "X-Requested-With": "XMLHttpRequest",
        }

    def _create_bbg_sef_payload(
        self,
        start_date: datetime,
        end_date: datetime,
        swap_type: Optional[Literal["getBsefEodIrsDataRequest", "getBsefEodFxDataRequest", "getBsefEodCdsDataRequest"]] = "getBsefEodIrsDataRequest",
    ):
        return {
            "Request": {
                swap_type: {
                    "tradeDays": {"startDay": start_date.strftime("%Y-%m-%dT%H:%M:%S.%fZ"), "endDay": end_date.strftime("%Y-%m-%dT%H:%M:%S.%fZ")}
                }
            }
        }

    async def _fetch_bbg_sef_data(
        self,
        client: httpx.AsyncClient,
        start_date: datetime,
        end_date: datetime,
        swap_type: Optional[Literal["getBsefEodIrsDataRequest", "getBsefEodFxDataRequest", "getBsefEodCdsDataRequest"]] = "getBsefEodIrsDataRequest",
        benchmark_rate_filter: Optional[str] = None,
        uid: Optional[int | str] = None,
        max_retries: Optional[int] = 3,
        backoff_factor: Optional[int] = 1,
    ):
        retries = 0
        try:
            while retries < max_retries:
                try:
                    response = await client.post(
                        self.bbg_sef_url,
                        json=self._create_bbg_sef_payload(start_date=start_date, end_date=end_date, swap_type=swap_type),
                        headers=self.bbg_sef_headers,
                        follow_redirects=False,
                        timeout=self._global_timeout,
                    )
                    response.raise_for_status()
                    json_data = response.json()
                    df = pd.DataFrame(json_data["response"]["getBsefEodIrsDataResponse"]["BsefEodData"])

                    if benchmark_rate_filter:
                        df = df[df["security"].str.contains(benchmark_rate_filter)]

                    if uid:
                        return df, uid
                    return df

                except httpx.HTTPStatusError as e:
                    self._logger.error(f"BBG SEF - Bad Status for {swap_type}-{start_date}-{end_date}: {response.status_code}")
                    if response.status_code == 404:
                        if uid:
                            return pd.DataFrame(), uid
                        return pd.DataFrame()

                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"BBG SEF - Throttled. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

                except Exception as e:
                    self._logger.error(f"BBG SEF - Error for {swap_type}-{start_date}-{end_date}: {e}")
                    print(e)
                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"BBG SEF - Throttled. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

            raise ValueError(f"BBG SEF - Max retries exceeded for {swap_type}-{start_date}-{end_date}")

        except Exception as e:
            self._logger.error(e)
            if uid:
                return pd.DataFrame(), uid
            return pd.DataFrame()

    async def _fetch_bbg_sef_data_semaphore(self, semaphore, *args, **kwargs):
        async with semaphore:
            return await self._fetch_bbg_sef_data(*args, **kwargs)

    def _get_business_days_groups(self, start_date: datetime, end_date: datetime, group_size=3):
        date_range = pd.date_range(start=start_date, end=end_date, freq="B")
        business_day_groups = [
            [bday.to_pydatetime() for bday in date_range[i : i + group_size].tolist()] for i in range(0, len(date_range), group_size)
        ]
        return business_day_groups

    def _swap_tenor_sort(self, column_name):
        parts = column_name.split()
        if len(parts) < 4:
            return (float("inf"), float("inf"))

        numeric_part = int(parts[-1][:-1])
        unit = parts[-1][-1]
        unit_order = {"D": 1, "W": 2, "M": 3, "Y": 4}
        return (unit_order.get(unit, float("inf")), numeric_part)

    def _liquid_tenors(self):
        return [
            "USD SWAP VS SOFR 1M",
            "USD SWAP VS SOFR 2M",
            "USD SWAP VS SOFR 3M",
            "USD SWAP VS SOFR 4M",
            "USD SWAP VS SOFR 6M",
            "USD SWAP VS SOFR 9M",
            "USD SWAP VS SOFR 1Y",
            "USD SWAP VS SOFR 18M",
            "USD SWAP VS SOFR 2Y",
            "USD SWAP VS SOFR 3Y",
            "USD SWAP VS SOFR 4Y",
            "USD SWAP VS SOFR 5Y",
            "USD SWAP VS SOFR 6Y",
            "USD SWAP VS SOFR 7Y",
            "USD SWAP VS SOFR 8Y",
            "USD SWAP VS SOFR 9Y",
            "USD SWAP VS SOFR 10Y",
            "USD SWAP VS SOFR 11Y",
            "USD SWAP VS SOFR 12Y",
            # "USD SWAP VS SOFR 13Y",
            # "USD SWAP VS SOFR 14Y",
            "USD SWAP VS SOFR 15Y",
            "USD SWAP VS SOFR 20Y",
            "USD SWAP VS SOFR 25Y",
            "USD SWAP VS SOFR 30Y",
            "USD SWAP VS SOFR 35Y",
            "USD SWAP VS SOFR 40Y",
            "USD SWAP VS SOFR 45Y",
            "USD SWAP VS SOFR 50Y",
        ]

    def _parse_usd_sofr_swap_tenor(self, tenor_str):
        match = re.match(r"USD SWAP VS SOFR (\d+)([MY])", tenor_str)
        if match:
            num, unit = match.groups()
            num = int(num)
            if unit == "M":
                return num / 12
            elif unit == "Y":
                return num
        return np.nan

    def _interpolate_row(self, row, times, interp_func):
        valid_cols = [col for col in row.index if pd.notna(row[col]) and pd.notna(times[col])]
        x = np.array([times[col] for col in valid_cols])
        y = np.array([row[col] for col in valid_cols])

        interp_cols = [col for col in row.index if pd.notna(times[col])]
        x_new = np.array([times[col] for col in interp_cols])

        if len(x) > 1:
            interpolated_rates = interp_func(x, y, x_new)
        else:
            interpolated_rates = np.full_like(x_new, np.nan)

        interpolated_series = pd.Series(interpolated_rates, index=interp_cols)
        for col in row.index:
            if pd.isna(times[col]):
                interpolated_series[col] = np.nan

        interpolated_series = interpolated_series.reindex(row.index)
        return interpolated_series

    def _interpolate_and_extrapolate_missing_rates(self, df, interp_func):
        times = {col: self._parse_usd_sofr_swap_tenor(col) for col in df.columns if col != "Date"}
        df_interpolated = df.apply(lambda row: self._interpolate_row(row, times, interp_func), axis=1)
        return df_interpolated

    def get_usd_sofr_swaps_historical_term_structures(
        self,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        max_concurrent_tasks: int = 64,
        max_keepalive_connections: int = 5,
        ohlc: Optional[Literal["open", "high", "low", "close"]] = None,
        tenors: Optional[List[str]] = None,
        liquid_tenors: Optional[bool] = False,
        interp_extrap_strat: Optional[Callable] = None,
    ) -> pd.DataFrame:
        groups = self._get_business_days_groups(start_date=start_date, end_date=end_date, group_size=30)
        start_end_date_tuples = [(min(group), max(group)) for group in groups]

        async def build_tasks(
            client: httpx.AsyncClient,
            start_end_date_tuples: List[Tuple[datetime, datetime]],
        ):
            tasks = []
            semaphore = asyncio.Semaphore(max_concurrent_tasks)
            for start_date, end_date in start_end_date_tuples:
                task = self._fetch_bbg_sef_data_semaphore(
                    semaphore=semaphore,
                    client=client,
                    start_date=start_date,
                    end_date=end_date,
                    swap_type="getBsefEodIrsDataRequest",
                    benchmark_rate_filter="USD SWAP VS SOFR",
                )
                tasks.append(task)

            return await asyncio.gather(*tasks)

        async def run_fetch_all(start_end_date_tuples: List[Tuple[datetime, datetime]]):
            limits = httpx.Limits(
                max_connections=max_concurrent_tasks,
                max_keepalive_connections=max_keepalive_connections,
            )
            async with httpx.AsyncClient(limits=limits) as client:
                all_data = await build_tasks(client=client, start_end_date_tuples=start_end_date_tuples)
                return all_data

        results: List[Tuple[str, pd.DataFrame]] = asyncio.run(run_fetch_all(start_end_date_tuples=start_end_date_tuples))
        big_df: pd.DataFrame = pd.concat(results)[["tradeDate", "security", f"price{ohlc.title()}" if ohlc else "settlementPrice"]]
        big_df = big_df[big_df["security"].str.len() <= 22]
        big_df["tradeDate"] = pd.to_datetime(big_df["tradeDate"], errors="coerce").dt.tz_localize(None)
        big_df.rename(columns={"tradeDate": "Date"}, inplace=True)
        big_df = big_df.pivot(index="Date", columns="security", values="settlementPrice")
        big_df = big_df[sorted(big_df.columns, key=self._swap_tenor_sort)]
        big_df.reset_index(inplace=True)

        if tenors:
            big_df = big_df[["Date"] + tenors]
        elif liquid_tenors:
            for tenor in self._liquid_tenors():
                if tenor not in big_df.columns:
                    big_df[tenor] = np.nan
            big_df = big_df[["Date"] + self._liquid_tenors()]

        if interp_extrap_strat is not None:
            date_col = big_df["Date"]
            big_df = self._interpolate_and_extrapolate_missing_rates(big_df[big_df.columns[1:]], interp_extrap_strat)
            big_df.insert(0, "Date", date_col)

        big_df.index.name = None
        big_df.columns.name = None

        return big_df

    def calc_sofr_swap_spreads(self, ct_yields_df: pd.DataFrame, sofr_swaps_df: pd.DataFrame) -> pd.DataFrame:
        common_dates = ct_yields_df.index.intersection(sofr_swaps_df.index)
        ct_yields_df = ct_yields_df.loc[common_dates]
        sofr_swaps_df = sofr_swaps_df.loc[common_dates]

        swaps_cols = [
            "USD SWAP VS SOFR 1M",
            "USD SWAP VS SOFR 2M",
            "USD SWAP VS SOFR 3M",
            "USD SWAP VS SOFR 4M",
            "USD SWAP VS SOFR 6M",
            "USD SWAP VS SOFR 1Y",
            "USD SWAP VS SOFR 2Y",
            "USD SWAP VS SOFR 3Y",
            "USD SWAP VS SOFR 5Y",
            "USD SWAP VS SOFR 7Y",
            "USD SWAP VS SOFR 10Y",
            "USD SWAP VS SOFR 20Y",
            "USD SWAP VS SOFR 30Y",
        ]
        ct_cols = ["CT1M", "CT2M", "CT3M", "CT4M", "CT6M", "CT1", "CT2", "CT3", "CT5", "CT7", "CT10", "CT20", "CT30"]

        spreads_df = sofr_swaps_df[swaps_cols].values - ct_yields_df[ct_cols].values
        spreads_df = pd.DataFrame(spreads_df, index=common_dates, columns=[f"{swp} Spread" for swp in swaps_cols])
        return spreads_df

    def get_usd_libor_swaps_historical_term_structures(
        self,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        max_concurrent_tasks: int = 64,
        max_keepalive_connections: int = 5,
        ohlc: Optional[Literal["open", "high", "low", "close"]] = None,
        tenors: Optional[List[str]] = None,
        liquid_tenors: Optional[bool] = False,
        interp_extrap_strat: Optional[Callable] = None,
    ):
        pass
